""" PyUAT based tracking """

import gzip
import tempfile
import time
import warnings
from pathlib import Path

from uatrack.config import setup_assignment_generators
from uatrack.core import simpleTracking
from uatrack.utils import extract_single_cell_information, save_tracking

from acia.attribute import attribute_tracking
from acia.base import ImageSequenceSource, Overlay
from acia.segm.formats import overlay_from_masks
from acia.tracking.formats import parse_simple_tracking
from acia.tracking.processor.utils import overlay_to_masks

from . import TrackingProcessor

TRACKING_CONFIGURATIONS = ["NN", "FO", "FO+G", "FO+O", "FO+DD", "FO+G+O+DD"]


class PyUATTracker(TrackingProcessor):
    """Processor for PyUAT: https://arxiv.org/abs/2503.21914"""

    def __init__(
        self,
        tracking_configuration: str,
        subsampling_factor=1,
        num_particles=1,
        num_cores=1,
        max_num_hypotheses=1,
        cutOff=-1,
        max_num_solutions=1,
        mip_method="CBC",
    ):
        """
        Args:
            tracking_configuration (str): PyUAT tracking configuration
        """

        if tracking_configuration not in TRACKING_CONFIGURATIONS:
            raise ValueError(
                f"'{tracking_configuration}' is not a valid tracking configuration. Pleas one of {TRACKING_CONFIGURATIONS}!"
            )

        self.config = tracking_configuration

        self.subsampling_factor = subsampling_factor

        self.num_particles = num_particles
        self.num_cores = num_cores
        self.max_num_hypotheses = max_num_hypotheses
        self.cutOff = cutOff
        self.max_num_solutions = max_num_solutions

        self.mip_method = mip_method

        if mip_method != "GRB":
            warnings.warn(
                "You are not using Gurobi! Please install gurobi and specify 'GRB' as optimizer for a tremendous speedup!"
            )

    def __call__(self, images: ImageSequenceSource, segmentation: Overlay):

        print("Extract single-cell information...")
        df, all_detections = extract_single_cell_information(segmentation)

        print("Setup assignment generators...")
        assignment_generators = setup_assignment_generators(
            df, self.subsampling_factor, self.config
        )

        print("Perform tracking...")
        # start tracking
        start = time.time()
        res = simpleTracking(
            df,
            assignment_generators,
            self.num_particles,
            num_cores=self.num_cores,
            max_num_hypotheses=self.max_num_hypotheses,
            cutOff=self.cutOff,
            max_num_solutions=self.max_num_solutions,
            mip_method=self.mip_method,  # use "GRB" if you have gurobi installed
        )
        end = time.time()

        print("time for tracking", end - start)

        with tempfile.TemporaryDirectory() as td:
            output_file = Path(td) / "simpleTracking.json.gz"
            save_tracking(res[0], all_detections, output_file)

            # read the tracking result
            with gzip.open(output_file) as input_file:
                tracking_overlay, tracking_graph = parse_simple_tracking(
                    input_file.read()
                )

        # Convert from contour based overlay to a mask based overlay
        height, width = images.get_frame(0).raw.shape[:2]
        mask_stack = overlay_to_masks(tracking_overlay, height, width)

        tracking_ov_new = overlay_from_masks(mask_stack)

        attribute_tracking(tracking_ov_new, None, tracking_graph, self)

        # TODO: create tracklet graph
        return tracking_ov_new, None, tracking_graph



"""LAP tracking processor: https://doi.org/10.1093/bioinformatics/btac799"""

import logging
import tempfile
from pathlib import Path

import networkx as nx
import numpy as np
import tifffile
from laptrack import OverLapTrack

from acia.attribute import attribute_tracking
from acia.base import ImageSequenceSource, Overlay
from acia.segm.formats import overlay_from_masks

from ..formats import read_ctc_tracking
from . import TrackingProcessor
from .utils import overlay_to_masks


class LAPTracker(TrackingProcessor):
    """Processor for LAP tracking"""

    def __init__(
        self,
        track_cost_cutoff=0.9,
        track_dist_metric_coefs=(1.0, -1.0, 0.0, 0.0, 0.0),
        gap_closing_dist_metric_coefs=(1.0, -1.0, 0.0, 0.0, 0.0),
        gap_closing_max_frame_count=1,
        splitting_cost_cutoff=0.9,
        splitting_dist_metric_coefs=(1.0, 0.0, 0.0, 0.0, -1.0),
    ):
        """Configure LAP tracker

        For the parameter configuration please refer to https://github.com/yfukai/laptrack/blob/main/docs/examples/overlap_tracking.ipynb
        """
        # define the overlap based tracking
        self.olt = OverLapTrack(
            track_cost_cutoff=track_cost_cutoff,
            track_dist_metric_coefs=track_dist_metric_coefs,
            gap_closing_dist_metric_coefs=gap_closing_dist_metric_coefs,
            gap_closing_max_frame_count=gap_closing_max_frame_count,
            splitting_cost_cutoff=splitting_cost_cutoff,
            splitting_dist_metric_coefs=splitting_dist_metric_coefs,
        )

    @staticmethod
    def __export(output_path, track_df, split_df, labels):
        output_path.mkdir(parents=True, exist_ok=True)

        df = track_df.reset_index()

        # relabel the cells in the segmentation masks
        new_labels = []

        for frame in np.unique(df["frame"]):
            frame_df = df[df["frame"] == frame]

            new_label = np.zeros_like(labels[frame])
            for _, row in frame_df.iterrows():

                label = row["label"]
                track_id = row["track_id"] + 1
                mask = labels[frame] == label
                new_label[mask] = track_id

            new_labels.append(new_label)

        # output the relabeled segmentation masks
        for i, new_label in enumerate(new_labels):
            tifffile.imwrite(
                output_path / f"t{i:04d}.tif", new_label, compression="zlib"
            )

        # Make the cell tracking format

        res = []

        for track_id in np.unique(df["track_id"]):
            frames = df[df["track_id"] == track_id]["frame"]
            start_frame = np.min(frames)
            last_frame = np.max(frames)

            if len(split_df) > 0:
                lookup = split_df[split_df["child_track_id"] == track_id]
            else:
                lookup = []

            if len(lookup) > 0:
                parent = lookup["parent_track_id"].item() + 1
            else:
                parent = 0

            res.append((track_id + 1, start_frame, last_frame, parent))

        with open(output_path / "man_track.txt", "w", encoding="utf-8") as csvfile:
            for row in res:
                csvfile.write(" ".join(map(str, row)) + "\n")

    def __call__(self, images: ImageSequenceSource, segmentation: Overlay):
        image = next(iter(images)).raw
        height, width = image.shape[:2]

        masks = overlay_to_masks(segmentation, height=height, width=width)

        if segmentation.numFrames() != len(masks):
            logging.warning("Number of segmented frames and masks is unequal!")

        track_df, split_df, _ = self.olt.predict_overlap_dataframe(masks)

        with tempfile.TemporaryDirectory() as td:
            self.__export(td, track_df, split_df, masks)

            ov, tracklet_graph, tracking_graph = read_ctc_tracking(td)

            attribute_tracking(ov, tracking_graph, tracking_graph, self)

            return ov, tracklet_graph, tracking_graph


class LAPTracker2(TrackingProcessor):
    """Processor for LAP tracking"""

    def __init__(
        self,
        laptrack_params,
    ):
        """Configure LAP tracker

        For the parameter configuration please refer to https://github.com/yfukai/laptrack/blob/main/docs/examples/overlap_tracking.ipynb
        """
        # define the overlap based tracking
        self.olt = OverLapTrack(**laptrack_params)

    @staticmethod
    def __export(output_path, track_df, split_df, labels):
        output_path.mkdir(parents=True, exist_ok=True)

        df = track_df.reset_index()

        # relabel the cells in the segmentation masks
        new_labels = []

        for frame in np.unique(df["frame"]):
            frame_df = df[df["frame"] == frame]

            new_label = np.zeros_like(labels[frame])
            for _, row in frame_df.iterrows():

                label = row["label"]
                track_id = row["track_id"] + 1
                mask = labels[frame] == label
                new_label[mask] = track_id

            new_labels.append(new_label)

        # output the relabeled segmentation masks
        for i, new_label in enumerate(new_labels):
            tifffile.imwrite(
                output_path / f"t{i:04d}.tif", new_label, compression="zlib"
            )

        # Make the cell tracking format

        res = []

        for track_id in np.unique(df["track_id"]):
            frames = df[df["track_id"] == track_id]["frame"]
            start_frame = np.min(frames)
            last_frame = np.max(frames)

            if len(split_df) > 0:
                lookup = split_df[split_df["child_track_id"] == track_id]
            else:
                lookup = []

            if len(lookup) > 0:
                parent = lookup["parent_track_id"].item() + 1
            else:
                parent = 0

            res.append((track_id + 1, start_frame, last_frame, parent))

        with open(output_path / "man_track.txt", "w", encoding="utf-8") as csvfile:
            for row in res:
                csvfile.write(" ".join(map(str, row)) + "\n")

    def __call__(self, images: ImageSequenceSource, segmentation: Overlay):
        image = next(iter(images)).raw
        height, width = image.shape[:2]

        masks = overlay_to_masks(segmentation, height=height, width=width)

        if segmentation.numFrames() != len(masks):
            logging.warning("Number of segmented frames and masks is unequal!")

        track_df, split_df, _ = self.olt.predict_overlap_dataframe(masks)

        with tempfile.TemporaryDirectory() as td:
            self.__export(Path(td), track_df, split_df, masks)

            ov, tracklet_graph, tracking_graph = read_ctc_tracking(td)

            return ov, tracklet_graph, tracking_graph


class LaptrackTracker(TrackingProcessor):
    """Processor for LAP tracking according to https://github.com/yfukai/laptrack/blob/main/docs/examples/overlap_tracking.ipynb"""

    def __init__(
        self,
        laptrack_params,
    ):
        """Configure LAP tracker

        For the parameter configuration please refer to https://github.com/yfukai/laptrack/blob/main/docs/examples/overlap_tracking.ipynb
        """
        # define the overlap based tracking
        self.olt = OverLapTrack(**laptrack_params)

    def __call__(self, images: ImageSequenceSource, segmentation: Overlay):
        image = next(iter(images)).raw
        height, width = image.shape[:2]

        mask_stack = overlay_to_masks(segmentation, height=height, width=width)

        track_df, split_df, _ = self.olt.predict_overlap_dataframe(mask_stack)

        tracking_ov = overlay_from_masks(mask_stack)

        label_lookup = {}
        for track_id, track_id_df in track_df.groupby("track_id"):
            label_lookup[track_id] = track_id_df.iloc[0].name[1]

        split_df["parent_track_label"] = split_df["parent_track_id"].apply(
            lambda id: label_lookup[id]
        )
        split_df["child_track_label"] = split_df["child_track_id"].apply(
            lambda id: label_lookup[id]
        )

        # create tracklet graph
        tracklet_graph = nx.DiGraph()

        for _, row in split_df.iterrows():
            tracklet_nodes = np.unique(track_df.reset_index()["label"])
            tracklet_graph.add_nodes_from(tracklet_nodes)
            tracklet_graph.add_edge(row["parent_track_label"], row["child_track_label"])

        # create tracking graph
        tracking_graph = nx.DiGraph()

        frame_label_lookup = {(cont.frame, cont.label): cont for cont in tracking_ov}

        track_sequences = {}

        for track_id, track_id_df in track_df.groupby("track_id"):
            track_seq = [(frame, label) for (frame, label), _ in track_id_df.iterrows()]
            label = track_id_df.iloc[0].name[1]
            track_sequences[label] = track_seq

        for label, track_seq in track_sequences.items():
            # add sequence
            for a, b in zip(track_seq[0:-1], track_seq[1:]):
                tracking_graph.add_edge(
                    frame_label_lookup[a].id, frame_label_lookup[b].id
                )

            # add divisions
            for parent_label in tracklet_graph.predecessors(label):
                # lookup the contours
                a = frame_label_lookup[track_sequences[parent_label][-1]].id
                b = frame_label_lookup[track_seq[0]].id

                print(f"{parent_label} -> {label}")
                print(f"{a} -> {b}")

                tracking_graph.add_edge(a, b)

        return tracking_ov, tracklet_graph, tracking_graph



"""Segmenter using YOLO: https://github.com/ultralytics/ultralytics"""

import logging

import numpy as np
from ultralytics import YOLO

from acia.attribute import attribute_segmentation
from acia.base import ImageSequenceSource, Instance, Overlay


class YOLOSegmenter:
    """YOLOSegmenter using Yolo: https://github.com/ultralytics/ultralytics"""

    def __init__(self, model):

        # create CellPose model
        self.model = YOLO(model)

    def __call__(self, images: ImageSequenceSource, conf=0.25, iou=0.7) -> Overlay:
        """Perform segmentation using yolo

        Args:
            images (ImageSequenceSource): the input image sequence
            conf (float, optional): Minimum confidence of detection objects. Defaults to 0.25.
            iou (float, optional): Objects with a higher IoU are supressed. Defaults to 0.7.

        Returns:
            Overlay: _description_
        """

        # list of images
        imgs = [im.raw for im in images]

        if len(imgs[0].shape) != 3 or imgs[0].shape[-1] != 3:
            logging.warning(
                "Wrong shape for YOLO images. They should have shape [H,W,3] but have %s",
                {imgs[0].shape},
            )

        # perform prediction using yolo
        results = self.model(imgs, retina_masks=True, conf=conf, iou=iou)

        # List of all instances
        instances = []

        # loop over all frames
        for frame, frame_data in enumerate(results):

            if frame_data.masks is None:
                # Nothing found within the image
                continue

            # loop over all masks
            for mask, box in zip(frame_data.masks, frame_data.boxes):

                # add one because they start couting at 0 (but zero is background in the mask)
                cls = int(box.cls.item()) + 1

                # get the mask data
                np_mask = (mask.data.cpu().numpy() * cls).astype(np.uint8)

                # create a new instance in the overlay
                instances.append(
                    Instance(np_mask[0], frame, cls, score=box.conf.item())
                )

        ov = Overlay(instances, frames=list(range(len(imgs))))

        attribute_segmentation(ov, self)

        return ov



from .output import CTCTrackingHelper
from . import TrackingSource
import networkx as nx

def life_cycle_lineage(tr_source: TrackingSource):

    overlay = tr_source.overlay
    tr_graph = tr_source.tracking_graph

    # compute the life-cycles of individual cells
    life_cycles = CTCTrackingHelper.compute_life_cycles(tr_graph)
    # create lookup (cont id --> life cycle index)
    life_cycle_lookup = CTCTrackingHelper.create_life_cycle_lookup(
        life_cycles
    )
    contour_lookup = {cont.id: cont for cont in overlay}

    lc_graph = nx.DiGraph()

    lc_graph.add_nodes_from(range(len(life_cycles)))

    for i in range(len(life_cycles)):
        lc_graph.nodes[i]["cycle"] = life_cycles[i]

    for lc_id, lc in enumerate(life_cycles):
        start = lc[0]

        parents = tr_graph.predecessors(start)

        for parent in parents:
            parent_lc_id = life_cycle_lookup[parent]
            
            lc_graph.add_edge(parent_lc_id, lc_id)

    for node in lc_graph:
        lc = lc_graph.nodes[node]["cycle"]

        lc_graph.nodes[node]["start_frame"] = contour_lookup[lc[0]].frame
        lc_graph.nodes[node]["end_frame"] = contour_lookup[lc[-1]].frame

    return lc_graph



import urllib
import zipfile
from pathlib import Path

def download_and_unzip(url: str, extract_dir: Path):

    zip_path, _ = urllib.request.urlretrieve(url)
    with zipfile.ZipFile(zip_path, "r") as f:
        f.extractall(extract_dir)


import random

# ----------------- TEST CASE ----------------

def build_lineage_large(seed=42, n_cells=32, max_branches=2, max_time=8):
    """
    Generate a random cell lineage graph for demo/testing.
    """
    random.seed(seed)
    G = nx.DiGraph()
    cell_count = 0
    time = 0
    alive = []
    for i in range(2):
        node = f"root{i}"
        G.add_node(node, t=0, label=f"Root{i}_L", color="red" if i==0 else "blue", val=i)
        alive.append(node)
        cell_count += 1
    for t in range(1, max_time+1):
        new_alive = []
        for cell in alive:
            n_daughters = random.choice([1, max_branches])
            for j in range(n_daughters):
                if cell_count >= n_cells:
                    break
                node = f"{cell}_{t}_{j}"
                node_label = node if random.random() > 0.2 else None
                extra = {"custom": f"extra_{node}", "val": random.randint(1, 10)}
                if node_label is not None:
                    G.add_node(node, t=t, label=node_label, **extra)
                else:
                    G.add_node(node, t=t, **extra)
                G.add_edge(cell, node)
                new_alive.append(node)
                cell_count += 1
        if cell_count < n_cells and random.random() < 0.4:
            node = f"new_{t}_{cell_count}"
            G.add_node(node, t=t, label=node, custom="from_new", val=-1)
            new_alive.append(node)
            cell_count += 1
        alive = new_alive
        if cell_count >= n_cells:
            break
    while G.number_of_nodes() < n_cells:
        t += 1
        node = f"late_{t}_{cell_count}"
        G.add_node(node, t=t, label=node, custom="late", val=-2)
        alive.append(node)
        cell_count += 1
    return G

if __name__ == '__main__':
    # Build test lineage graph
    G = build_lineage_large(n_cells=32, max_branches=2, max_time=10)

    # --- Matplotlib plot (static) ---
    plt.figure(figsize=(14, 7))
    plot_cell_lineage(
        G, orientation='horizontal',
        show_label=True, label_name='label',
        node_marker="o", node_ms=10,
        line_color="forestgreen", line_lw=3,
        mark_births=True, birth_color="lime", birth_marker=">", birth_ms=18,
        mark_ends=True, end_color="magenta", end_marker='s', end_ms=18,
        interactive_tooltip=True  # Enable tooltips if mplcursors installed
    )
    plt.title("Matplotlib: node features on hover (try with mplcursors)")
    plt.show()

    # --- Plotly interactive plot ---
    import plotly
    fig = plotly_cell_lineage(
        G, orientation='horizontal',
        show_label=True, label_name='label',
        node_marker="circle", node_ms=14,
        line_color="forestgreen", line_width=3,
        mark_births=True, birth_color="lime", birth_marker=">", birth_ms=20,
        mark_ends=True, end_color="magenta", end_marker='s', end_ms=20,
        figure_title="Plotly: node features as pseudo-table in hover"
    )
    fig.show()



""" Ultrack based tracking """

import networkx as nx
from ultrack import to_tracks_layer, track, tracks_to_zarr
from ultrack.utils import labels_to_contours

from acia.attribute import attribute_tracking
from acia.base import ImageSequenceSource, Overlay
from acia.segm.formats import overlay_from_masks

from . import TrackingProcessor
from .utils import overlay_to_masks


class UltrackTracker(TrackingProcessor):
    """Processor for Ultrack: https://github.com/royerlab/ultrack/tree/main"""

    def __init__(self, config):
        """_summary_

        Args:
            config (_type_): Ultrack configuration
        """

        self.config = config

    def __call__(self, images: ImageSequenceSource, segmentation: Overlay):

        image = next(iter(images)).raw
        height, width = image.shape[:2]

        mask_stack = overlay_to_masks(segmentation, height=height, width=width)

        detection, edges = labels_to_contours(mask_stack, sigma=4.0)

        # perform the tracking
        track(
            foreground=detection,
            edges=edges,
            config=self.config,
            overwrite=True,
        )

        # convert back
        tracks_df, graph = to_tracks_layer(self.config)
        labels = tracks_to_zarr(self.config, tracks_df)

        # parse the tracking overlay from instance stack (TxHxW)
        tracking_ov = overlay_from_masks(labels)

        # create the tracklet graph
        tracklet_graph = nx.DiGraph()
        for child, parent in graph.items():
            tracklet_graph.add_edge(parent, child)

        # compute tracking graph

        label_contours = {}
        for cont in tracking_ov:
            label_contours[cont.label] = label_contours.get(cont.label, []) + [cont]

        tracking_graph = nx.DiGraph()

        for label, contours in label_contours.items():
            sorted(contours, key=lambda c: c.frame)
            # print([c.frame for c in contours])

            tracklet_graph.nodes[label]["start_frame"] = contours[0].frame
            tracklet_graph.nodes[label]["end_frame"] = contours[-1].frame

            tracking_graph.add_nodes_from([c.id for c in contours])

            # add sequential tracks
            for c1, c2 in zip(contours[:-1], contours[1:]):
                tracking_graph.add_edge(c1.id, c2.id)

            # add divisions
            parent_labels = list(tracklet_graph.predecessors(label))

            # add connection to every parent
            for pl in parent_labels:
                last_cont = label_contours[pl][-1]
                tracking_graph.add_edge(last_cont.id, contours[0].id)

        attribute_tracking(tracking_ov, tracklet_graph, tracking_graph, self)

        return tracking_ov, tracklet_graph, tracking_graph



""" Trackastra based tracking """

import logging
import tempfile
from pathlib import Path

import numpy as np
import torch
from trackastra.model import Trackastra
from trackastra.tracking import graph_to_ctc

from acia.attribute import attribute_tracking
from acia.base import ImageSequenceSource, Overlay
from acia.segm.formats import read_ctc_segmentation_native
from acia.tracking import ctc_track_graph
from acia.tracking.formats import read_ctc_tracklet_graph

from . import TrackingProcessor
from .utils import overlay_to_masks


class TrackastraTracker(TrackingProcessor):
    """Processor for Trackastra: https://doi.org/10.48550/arXiv.2405.15700"""

    def __init__(self, mode="greedy"):

        device = "cuda" if torch.cuda.is_available() else "cpu"

        # Load a pretrained model
        self.model = Trackastra.from_pretrained("general_2d", device=device)
        self.mode = mode

    def __call__(self, images: ImageSequenceSource, segmentation: Overlay):

        image = next(iter(images)).raw
        height, width = image.shape[:2]

        masks = overlay_to_masks(segmentation, height=height, width=width)
        imgs = np.stack([im.raw for im in images])

        if len(imgs.shape) == 4:
            # strip off last because it should not be used
            imgs = imgs[..., 0]

        if segmentation.numFrames() != len(masks):
            logging.warning("Number of segmented frames and masks is unequal!")

        # perform the actual tracking
        track_graph = self.model.track(imgs, masks, mode=self.mode)

        # Write to cell tracking challenge format
        with tempfile.TemporaryDirectory() as td:
            _, _ = graph_to_ctc(
                track_graph,
                masks,
                outdir=td,
            )

            input_path = Path(td)
            track_file = input_path / "man_track.txt"

            ov = read_ctc_segmentation_native(input_path)
            tracklet_graph = read_ctc_tracklet_graph(track_file)

        tracking_graph = ctc_track_graph(ov, tracklet_graph)

        attribute_tracking(ov, tracklet_graph, tracking_graph, self)

        return ov, tracklet_graph, tracking_graph



""" Module to convert tracking formats """

from __future__ import annotations

import json
import logging
from pathlib import Path

import networkx as nx
import numpy as np
import pandas as pd
import tifffile

from acia.base import Contour, ImageSequenceSource, Overlay
from acia.segm.formats import read_ctc_segmentation_native


def parse_simple_tracking(file_content: str) -> tuple[Overlay, nx.DiGraph]:
    """Parse simple tracking format from file content string

    Args:
        file_content (str): simple tracking format file content

    Returns:
        Tuple[Overlay, nx.DiGraph]: segmentation overlay and tracking graph
    """

    data = json.loads(file_content)

    segmentation_data = data["segmentation"]
    tracking_data = data["tracking"]

    # deal with the segmentation first
    all_detections = []

    # create contours
    for det in segmentation_data:

        det_id = det["id"]
        # try to convert to integer
        try:
            det_id = int(det_id)
        except ValueError:
            pass

        all_detections.append(
            Contour(det["contour"], -1, det["frame"], det_id, det["label"])
        )

    segmentation_overlay = Overlay(all_detections)

    # deal with the tracking
    tracking = nx.DiGraph()

    tracking.add_nodes_from(map(lambda cont: cont.id, segmentation_overlay))
    node_set = set(tracking.nodes)
    for cont in segmentation_overlay:
        tracking.nodes[cont.id]["frame"] = cont.frame

    # create graph from id links
    for link in tracking_data:
        if link["sourceId"] in node_set and link["targetId"] in node_set:
            tracking.add_edge(link["sourceId"], link["targetId"])

    return segmentation_overlay, tracking


def gen_simple_tracking(overlay: Overlay, tracking_graph: nx.Graph) -> str:
    """Create a simple tracking format from overlay and tracking graph

    Args:
        overlay (Overlay): segmentation overlay
        tracking_graph (nx.Graph): tracking graph

    Returns:
        str: simple tracking format string
    """

    segmentation_data = []
    for cont in overlay:

        coordinates = cont.coordinates

        if isinstance(coordinates, np.ndarray):
            coordinates = coordinates.tolist()

        segmentation_data.append(
            dict(label=cont.label, contour=coordinates, id=cont.id, frame=cont.frame)
        )

    tracking_data = []
    for edge in tracking_graph.edges:
        tracking_data.append(dict(sourceId=edge[0], targetId=edge[1]))

    simpleTracking = dict(segmentation=segmentation_data, tracking=tracking_data)

    return json.dumps(simpleTracking)


def read_ctc_tracklet_graph(file: Path):
    colnames = ["label", "t_start", "t_end", "parent"]

    ctc_df = pd.read_csv(
        file,
        names=colnames,
        header=None,
        dtype={"label": "int", "t_start": "int", "t_end": "int", "parent": "int"},
        sep=" ",
    )

    tracklet_graph = nx.DiGraph()

    for _, row in ctc_df.iterrows():
        label, t_start, t_end, parent_label = row.to_list()
        tracklet_graph.add_node(label, start_frame=t_start, end_frame=t_end)

        if parent_label != 0:
            tracklet_graph.add_edge(parent_label, label)

    return tracklet_graph


def read_ctc_tracking(input_path: Path) -> tuple[Overlay, nx.DiGraph, nx.DiGraph]:
    """Read ctc tracking information

    Args:
        input_path (Path): Path to the ctc tracking folder

    Returns:
        tuple[Overlay, nx.DiGraph, nx.DiGraph]: segmentation overlay, tracklet graph (every cell cycle is a node), tracking graph (every cell detection is a node)
    """
    input_path = Path(input_path)
    track_file = input_path / "man_track.txt"

    ov = read_ctc_segmentation_native(input_path)
    tracklet_graph = read_ctc_tracklet_graph(track_file)

    tracking_graph = ctc_track_graph(ov, tracklet_graph)

    return ov, tracklet_graph, tracking_graph


def write_ctc_tracking(
    output_path: Path,
    images: ImageSequenceSource,
    overlay: Overlay,
    tracklet_graph: nx.DiGraph,
):
    """Write ctc tracking to output folder

    Args:
        output_path (Path): output folder for writing
        images (ImageSequenceSource): image time-lapse (only used to compute mask sizes)
        overlay (Overlay): segmentation overlay
        tracklet_graph (nx.DiGraph): tracklet graph (every cell cycle is a node)
    """

    output_path = Path(output_path)

    # Write tracklet information
    data = []
    for n in tracklet_graph.nodes:

        predecessors = list(tracklet_graph.predecessors(n))
        if len(predecessors) == 0:
            parent_label = 0
        else:
            parent_label = predecessors[0]

            if len(predecessors) > 1:
                logging.warning(
                    "Tracklet has more than one parent. This indicates a malformed tracklet graph!"
                )

        data.append(
            {
                "label": n,
                "t_start": tracklet_graph.nodes[n]["start_frame"],
                "t_end": tracklet_graph.nodes[n]["end_frame"],
                "parent": parent_label,
            }
        )

    df_ctc = pd.DataFrame(data)
    df_ctc.to_csv(output_path / "man_track.txt", sep=" ", header=False, index=False)

    # get the image size
    height, width = next(iter(images)).raw.shape[:2]

    # Write segmentation information
    for i, frame_overlay in enumerate(overlay.timeIterator()):
        local_mask = np.zeros((height, width), dtype=np.uint16)
        for cont in frame_overlay:
            cont_mask = cont.toMask(height, width).astype(np.uint16) * cont.label
            local_mask = np.maximum(local_mask, cont_mask)

        tifffile.imwrite(output_path / f"man_track{i:04d}.tif", local_mask)


def ctc_track_graph(ov: Overlay, tracklet_graph: nx.DiGraph):
    """Computes the ctc track graph (every cell detection is a node) based on cell detections (overlay) and the tracklet graph (every tracklet is one node).

    Hint: overlay labels and tracklet_graph node ids need to align.

    Args:
        ov (Overlay): _description_
        tracklet_graph (nx.DiGraph): _description_

    Returns:
        _type_: _description_
    """

    track_graph = nx.DiGraph()

    # add all the nodes
    for cont in ov:
        track_graph.add_node(cont.id, frame=cont.frame)

    tracklets = {}
    for cont in ov:
        tracklets[cont.label] = tracklets.get(cont.label, []) + [cont]

    for tracklet_label in tracklets:
        tracklets[tracklet_label] = sorted(
            tracklets[tracklet_label], key=lambda c: c.frame
        )

    for tracklet_label, tracklet_nodes in tracklets.items():

        # add tracklet edges
        for contA, contB in zip(tracklet_nodes, tracklet_nodes[1:]):
            track_graph.add_edge(contA.id, contB.id)

        for pred_label in tracklet_graph.predecessors(tracklet_label):
            track_graph.add_edge(tracklets[pred_label][-1].id, tracklet_nodes[0].id)

        for succ_label in tracklet_graph.successors(tracklet_label):
            track_graph.add_edge(tracklet_nodes[-1].id, tracklets[succ_label][0].id)

    return track_graph


def tracking_to_graph(data: list[dict]) -> nx.DiGraph:
    """Populates a ctc tracking into a full tracking lineage where every detection has its own node with a unique id based on (ctc_id, frame)

    Args:
        data (list[dict]): Output of :func:`read_ctc_tracking`

    Returns:
        nx.DiGraph: A lineage graph where every detection has its unique node (id, frame) and the edges represent the linking
    """

    graph = nx.DiGraph()

    # go through every ctc line
    for item in data:
        # iterate every frame for that the cell track exists
        for frame in range(int(item["start_frame"]), int(item["end_frame"]) + 1):
            # add node with unique id (ctc_id, frame)
            graph.add_node((item["id"], frame), frame=frame)
            # add non-division links
            if graph.has_node((item["id"], frame - 1)):
                graph.add_edge((item["id"], frame - 1), (item["id"], frame))

    # add division links
    for item in data:
        if int(item["parent_id"]) != 0:

            # get the time it divides
            # split_frame = int(item["start_frame"]) - 1

            # extract source and target
            source_candidates = list(
                # pylint: disable=cell-var-from-loop
                filter(lambda n: n[0] == item["parent_id"], graph.nodes)
            )
            latest_source = np.argmax(list(map(lambda n: n[1], source_candidates)))
            source = source_candidates[
                latest_source
            ]  # (item["parent_id"], split_frame)
            target = (item["id"], int(item["start_frame"]))

            if not graph.has_node(source):
                print("Error")

            assert graph.has_node(source), f"Source: {source}"
            assert graph.has_node(target), f"Target: {target}"

            # add the edge
            graph.add_edge(source, target)

    return graph



"""Top-level package for AutomatedCellularImageAnalysis."""

__author__ = """Johannes Seiffarth"""
__email__ = "j.seiffarth@fz-juelich.de"
__version__ = "0.3.1"


import pint

ureg = pint.get_application_registry()
Q_ = ureg.Quantity
U_ = ureg.Unit



"""Module for testing visualization utils module"""

import unittest
from datetime import timedelta

from acia.viz.utils import strfdelta


class TestUtils(unittest.TestCase):
    """Testing scalebar and time overlay"""

    def test_formatting(self):
        # Check that video file exists
        self.assertEqual(
            strfdelta(timedelta(days=2, hours=12, minutes=22, seconds=13)),
            "02d 12h 22m 13s",
        )


if __name__ == "__main__":
    unittest.main()



"""Attribute usage"""

import logging
import os
from datetime import datetime, timezone

import networkx as nx
from influxdb_client import InfluxDBClient, Point, WritePrecision
from influxdb_client.client.write_api import SYNCHRONOUS

from acia.base import Overlay

logger = logging.getLogger(__name__)

TOKEN = "58M8R17lQgh-U17snMPjya139MFD9_de2KqseKmJihcAZ6mkohozd-9-rvz71SZLI7GI1Qz8Nii2QxI1QjcVkw=="
URL = "https://worldofmicrobes.ddnss.org:8086"
ORG = "com.microbes"

#### Attribute functionality
def attribute_segmentation(overlay: Overlay, segmentation_processor=None):
    # pylint: disable=broad-except
    try:
        user = os.environ.get(
            "JUPYTERHUB_USER", os.environ.get("USER", os.environ.get("USERNAME"))
        )
        bucket = "segmentation"

        # initialize client from environment properties
        with InfluxDBClient(url=URL, token=TOKEN) as client:
            # setup write api
            write_api = client.write_api(write_options=SYNCHRONOUS)

            # get current time
            now = datetime.now(timezone.utc)

            point = Point("segmentation")

            point.tag("user", user)
            if segmentation_processor is not None:
                point.tag("method", segmentation_processor.__class__)

            # add the count field
            point.field("num_cells", len(overlay))
            point.field("num_images", len(overlay.frames()))

            # set the time
            point.time(now, WritePrecision.S)

            # write (will already be batched)
            write_api.write(bucket=bucket, record=point, org=ORG)
    except Exception as e:
        logger.exception(e)


def attribute_tracking(
    tracking_ov: Overlay,
    tracklet_graph: nx.DiGraph,
    tracking_graph: nx.DiGraph,
    tracker=None,
):
    # pylint: disable=broad-except
    try:
        user = os.environ.get(
            "JUPYTERHUB_USER", os.environ.get("USER", os.environ.get("USERNAME"))
        )
        bucket = "tracking"

        # initialize client from environment properties
        with InfluxDBClient(url=URL, token=TOKEN) as client:
            # setup write api
            write_api = client.write_api(write_options=SYNCHRONOUS)

            # get current time
            now = datetime.now(timezone.utc)

            point = Point("tracking")

            point.tag("user", user)
            if tracker is not None:
                point.tag("method", tracker.__class__)

            # add the count field
            point.field("num_cells", len(tracking_ov))
            point.field("tracking_nodes", tracking_graph.number_of_nodes())
            point.field("tracking_edges", tracking_graph.number_of_edges())
            if tracklet_graph is not None:
                point.field("tracklets_nodes", tracklet_graph.number_of_nodes())
                point.field("tracklets_edges", tracklet_graph.number_of_edges())
            point.field("num_images", len(tracking_ov.frames()))

            # set the time
            point.time(now, WritePrecision.S)

            # write (will already be batched)
            write_api.write(bucket=bucket, record=point, org=ORG)
    except Exception as e:
        logger.exception(e)



"""Segmentation Processors"""

from acia.base import ImageSequenceSource, Overlay


class SegmentationProcessor:
    """Base class for segmentation processors"""

    def __call__(self, images: ImageSequenceSource) -> Overlay:
        raise NotImplementedError("Please implement this base function")



"""Segmenter using Cellpose (below v4)"""

from cellpose import core, models

from acia.attribute import attribute_segmentation
from acia.base import ImageSequenceSource, Overlay
from acia.segm.formats import overlay_from_masks


class CellposeSegmenter:
    """Cellpose segmenter"""

    def __init__(self, model_type: str):
        self.use_GPU = core.use_gpu()
        self.model = models.CellposeModel(gpu=self.use_GPU, model_type=model_type)

    @staticmethod
    def __predict(images, model, cellpose_params=None):
        chans = [0, 0]  # this means segment based on first channel, no second channel

        if cellpose_params is None:
            cellpose_params = {}

        masks, _, _ = model.eval(images, channels=chans, **cellpose_params)

        return masks

    def __call__(self, images: ImageSequenceSource, cellpose_params=None) -> Overlay:

        imgs = []
        for image in images:
            raw_image = image.raw

            imgs.append(raw_image)

        masks = self.__predict(imgs, self.model, cellpose_params=cellpose_params)

        ov = overlay_from_masks(masks)

        attribute_segmentation(ov, self)

        return ov



"""Omnipose segmentation implementation"""

from pathlib import Path

import numpy as np
import torch
from cellpose_omni import models
from tqdm.auto import tqdm

from acia.attribute import attribute_segmentation
from acia.base import ImageSequenceSource, Overlay
from acia.segm.formats import overlay_from_masks

from . import SegmentationProcessor


def batch(iterable, n=1):
    l = len(iterable)
    for ndx in range(0, l, n):
        yield iterable[ndx : min(ndx + n, l)]


class OmniposeSegmenter(SegmentationProcessor):
    """Omnipose segmentation implementation"""

    def __init__(self, use_GPU: bool = None, model="bact_phase_omni"):
        if use_GPU is None:
            use_GPU = torch.cuda.is_available()
        self.use_GPU = use_GPU

        model_type = None
        model_path = None

        if Path(model).exists() and Path(model).is_file():
            model_path = model
        elif model in models.MODEL_NAMES:
            model_type = model
        else:
            raise ValueError(
                "Specified model is neither predefined nor a url to download"
            )

        if model_type:
            self.model = models.CellposeModel(gpu=use_GPU, model_type=model_type)
        if model_path:
            self.model = models.CellposeModel(
                gpu=use_GPU, pretrained_model=model_path, nclasses=3, nchan=2
            )

    @staticmethod
    def __predict(images, model, omnipose_parameters: dict = None, batch_size=20):

        if omnipose_parameters is None:
            omnipose_parameters = {}

        chans = [0, 0]  # this means segment based on first channel, no second channel

        # define parameters
        mask_threshold = -1
        verbose = 0  # turn on if you want to see more output
        transparency = True  # transparency in flow output
        rescale = (
            None  # give this a number if you need to upscale or downscale your images
        )
        omni = True  # we can turn off Omnipose mask reconstruction, not advised
        flow_threshold = 0.4  # default is .4, but only needed if there are spurious masks to clean up; slows down output
        resample = (
            True  # whether or not to run dynamics on rescaled grid or original grid
        )
        cluster = True  # use DBSCAN clustering

        all_masks = []

        pbar = tqdm(
            total=len(images),
            desc="Batched Omnipose prediction...",
        )

        for image_batch in batch(images, n=batch_size):

            # Make evaluation (flows and styles are not needed)
            masks, _, _ = model.eval(
                image_batch,
                channels=chans,
                rescale=rescale,
                mask_threshold=mask_threshold,
                transparency=transparency,
                flow_threshold=flow_threshold,
                omni=omni,
                cluster=cluster,
                resample=resample,
                verbose=verbose,
                model_loaded=True,
                show_progress=False,
                **omnipose_parameters,
            )

            all_masks.append(masks)

            pbar.update(len(image_batch))

        return np.concatenate(all_masks)

    def predict(self, images: ImageSequenceSource) -> Overlay:
        return self(images)

    def __call__(
        self, images: ImageSequenceSource, omnipose_parameters: dict = None
    ) -> Overlay:

        imgs = []
        for image in images:
            raw_image = image.raw

            # Reduce HxWxC=1 image to HxW shape
            if len(raw_image.shape) == 3:
                if raw_image.shape[2] != 1:
                    raise ValueError(
                        f"Omnipose Segmenter only accepts a single channel image. Currently it is HxWxC: {raw_image.shape}"
                    )

                # make it a grayscale image
                raw_image = raw_image[..., 0]

            imgs.append(raw_image)

        masks = self.__predict(
            imgs, self.model, omnipose_parameters=omnipose_parameters
        )

        ov = overlay_from_masks(masks)

        attribute_segmentation(ov, self)

        return ov



"""Utils for segmentation data handling"""

from typing import Tuple

import numpy as np
from shapely.geometry import MultiPolygon, Polygon
from shapely.ops import unary_union
from tqdm.auto import tqdm

from acia.base import Contour, Overlay
from acia.utils import pairwise_distances


def compute_indices(frame: int, size_t: int, size_z: int) -> Tuple[int, int]:
    """Compute t and z values from a linearized frame number

    Args:
        frame (int): the linearized frame index
        size_t (int): the total size of the t dimension
        size_z (int): the total size of the z dimension

    Returns:
        (Tuple[int, int]): tuple of (t,z) indices
    """

    if size_t > 1 and size_z > 1:
        t = int(np.floor(frame / size_t))
        z = frame % size_t
    elif size_t > 1:
        t = frame
        z = 0
    elif size_z >= 1:
        t = 0
        z = frame
    elif size_t == 1 and size_z == 1:
        t = 0
        z = 0
    else:
        raise ValueError("This state should not be reachable!")

    return t, z


def length_and_area(contour: Contour) -> Tuple[float, float]:
    """Compute length and area of a contour object (in pixel coordinates)

    Args:
        contour (Contour): contour object

    Returns:
        tuple[float, float]: length and area of the contour
    """

    polygon = Polygon(contour.coordinates)

    length = np.max(
        pairwise_distances(np.array(polygon.minimum_rotated_rectangle.exterior.coords))
    )
    return length, polygon.area


def merge_cells_to_colonies(overlay: Overlay, expand=10) -> Overlay:
    """Computing colony blobs from single-cell overlay

    Args:
        overlay (Overlay): Single-cell overlay containing the individual cell objects
        expand (int, optional): The number of pixels to expand single-cell objects in order to form blobs. Defaults to 10.

    Returns:
        Overlay: Overlay of colony blobs
    """

    merged_contours = []

    # iterate over frames and all the cell instances
    for frame_ind, frame_overlay in enumerate(
        tqdm(overlay.timeIterator(), desc="Merging cells to colonies...")
    ):

        # get all polygons
        cont_polys = [cont.polygon for cont in frame_overlay]

        # increase their size (like a dilation)
        oversized_polys = [poly.buffer(expand) for poly in cont_polys]

        # merge all polys
        intersection = unary_union(oversized_polys)

        # erose the merged polygon
        i = intersection.buffer(-expand)

        # make it a contour
        polygons = [i]
        if isinstance(i, MultiPolygon):
            polygons = list(i.geoms)

        contours = [
            Contour(
                np.array(list(zip(p.exterior.xy))).T.squeeze(), -1, frame_ind, frame_ind
            )
            for p in polygons
        ]

        contours = list(filter(lambda c: len(c.coordinates) >= 3, contours))

        # add merged contour to results
        merged_contours += contours

    # return new overlay with merged contours
    return Overlay(merged_contours)



""" Test acia base functionality"""

import unittest

import numpy as np

from acia.base import Contour, Instance


class TestContour(unittest.TestCase):
    """Test contour functionality"""

    def test_center(self):
        contour = [[0, 0], [1, 0], [1, 1], [0, 1]]

        self.assertTrue(contour is not None)

        # simple contour
        np.testing.assert_array_equal(
            Contour(contour, 0, 0, 0).center, np.array([0.5, 0.5], dtype=np.float32)
        )

        # unequal point sampling
        contour = [[0, 0], [0.5, 0], [1, 0], [1, 1], [0, 1]]
        np.testing.assert_array_equal(
            Contour(contour, 0, 0, 0).center, np.array([0.5, 0.5], dtype=np.float32)
        )

    def test_rasterization(self):
        """Make sure that contour to mask rasterization preserves area"""

        contours = [[[0, 0], [10, 0], [10, 10], [0, 10]]]

        for coordinates in contours:
            cont = Contour(coordinates, -1, 0, -1)
            mask = cont.toMask(40, 40)

            self.assertEqual(cont.area, np.sum(mask))


class TestInstance(unittest.TestCase):
    """Test contour functionality"""

    def test_center(self):
        mask = np.array(
            [
                [0, 0, 0, 0, 0],
                [0, 0, 3, 3, 0],
                [0, 3, 3, 3, 3],
                [0, 2, 2, 0, 0],
            ],
            dtype=np.uint8,
        )

        instance = Instance(mask, 0, 3)

        self.assertEqual(instance.area, 6)

        poly = instance.polygon

        self.assertEqual(poly.area, 6)

        instance = Instance(mask, 0, 2)
        self.assertEqual(instance.area, 2)

        poly = instance.polygon
        self.assertEqual(poly.area, 2)


if __name__ == "__main__":
    unittest.main()



import timeit
import numpy as np
from acia.viz import VideoExporter2, VideoExporter

def run_encoder(encoder):
    images = np.random.randint(0, 255, size=(100, 1024, 1024, 3), dtype=np.uint8)

    with encoder as ve:
        for image in images:
            ve.write(image)


def func():
    run_encoder(VideoExporter2.default_h265("test.mp4", 3))

def func_old():
    run_encoder(VideoExporter("test.mp4", 3, codec="vp09"))

run_encoder(VideoExporter2.default_av1("test.mkv", 3))

#print(timeit.repeat(func, number=1, repeat=1))
#print(timeit.repeat(func_old, number=1, repeat=1))


import unittest

from acia.segm.omero.storer import OmeroSequenceSource
import time

class TestIndexing(unittest.TestCase):
    """Test the linearization of z and t stacks"""

    def test_conn_loss(self):
        iss = OmeroSequenceSource(978, username="jseiffarth", password="jseiffarth", serverUrl="localhost")

        image = next(iter(iss))

        time.sleep(15*60)
        #iss.conn.close()

        other_image = next(iter(iss))

if __name__ == "__main__":
    unittest.main()



"""Module for testing visualization module for lineage rendering"""

import unittest
import networkx as nx

from acia.viz import hierarchy_pos_loop_multi, plot_lineage_tree

# Example usage (using previous forest creation and layout code)
def build_forest_example():
    G = nx.DiGraph()
    G.add_edge('A0', 'A1')
    G.add_edge('A1', 'A2')
    G.add_edge('A2', 'A3')
    G.add_edge('B0', 'B1')
    G.add_edge('B0', 'B2')
    G.add_edge('B1', 'B3')
    G.add_edge('B1', 'B4')
    G.add_edge('C0', 'C1')
    G.add_edge('C0', 'C2')
    G.add_edge('C0', 'C3')
    G.add_edge('C3', 'C4')
    G.add_edge('C4', 'C5')

    # Add y property (demo: last character as integer)
    for node in G.nodes():
        try:
            G.nodes[node]['y'] = int(node[-1])
        except:
            G.nodes[node]['y'] = 0
    return G


class TestLineageRendering(unittest.TestCase):
    """Testing scalebar and time overlay"""

    def test_formatting(self):
        # Build and plot the forest
        G = build_forest_example()
        roots = [n for n, d in G.in_degree() if d == 0]
        pos = hierarchy_pos_loop_multi(G, roots)
        plot_lineage_tree(G, pos, mode='vertical', draw_labels=True, title="Vertical with labels", y_attr="y")
        plot_lineage_tree(G, pos, mode='vertical', flip_vertical=True, draw_labels=True, title="Vertical with labels (flipped)")
        plot_lineage_tree(G, pos, mode='horizontal', draw_labels=True, title="Horizontal with labels")
        plot_lineage_tree(G, pos, mode='horizontal', flip_horizontal=True, draw_labels=True, title="Horizontal flipped with labels")

    def test_imbalance(self):
        # --- Build the imbalanced forest ---
        G = nx.DiGraph()
        # First tree: chain
        G.add_edge('R1', 'R1-1')
        G.add_edge('R1-1', 'R1-2')
        # Second tree: branch
        G.add_edge('R2', 'R2-1')
        G.add_edge('R2', 'R2-2')
        # Set custom timepoints
        G.nodes['R1']['y'] = 0
        G.nodes['R1-1']['y'] = 1
        G.nodes['R1-2']['y'] = 2
        G.nodes['R2']['y'] = 2
        G.nodes['R2-1']['y'] = 3
        G.nodes['R2-2']['y'] = 3

        roots = [n for n, d in G.in_degree() if d == 0]
        pos = hierarchy_pos_loop_multi(G, roots)
        plot_lineage_tree(
            G, pos, mode='vertical', draw_labels=True,
            title="Imbalanced forest with multiple roots at different y", y_attr='y'
        )



if __name__ == "__main__":
    unittest.main()



"""Utilities for tracking
"""

import logging
from itertools import product

import networkx as nx
import numpy as np
from shapely.geometry import MultiPolygon
from shapely.ops import unary_union

from acia.base import Contour, Instance, Overlay
from acia.tracking.output import CTCTrackingHelper


def life_cycle_lineage(tr_graph: nx.DiGraph) -> nx.DiGraph:
    """Compresses populated lineage to life cycle lineage (one node per cell cycle)

    Args:
        tr_graph (nx.DiGraph): populated tracking graph

    Returns:
        nx.DiGraph: Life cycle lineage with cell cylces as nodes
    """

    # compute the life-cycles of individual cells
    life_cycles = CTCTrackingHelper.compute_life_cycles(tr_graph)
    # create lookup (cont id --> life cycle index)
    life_cycle_lookup = CTCTrackingHelper.create_life_cycle_lookup(life_cycles)
    # contour_lookup = {cont.id: cont for cont in overlay}

    lc_graph = nx.DiGraph()

    # add all the nodes
    lc_graph.add_nodes_from(range(len(life_cycles)))

    # set the "cycle" property to contain the populated life cycle nodes
    for i, life_cycle in enumerate(life_cycles):
        lc_graph.nodes[i]["cycle"] = life_cycle

    # iterate over life_cycles
    for lc_id, lc in enumerate(life_cycles):
        start = lc[0]

        # extract parents from populated tracking
        parents = tr_graph.predecessors(start)

        for parent in parents:
            # get the parent life_cycle
            parent_lc_id = life_cycle_lookup[parent]

            # establish an edge between parent and child
            lc_graph.add_edge(parent_lc_id, lc_id)

    # set "start_frame" and "end_frame" for every node in the life cycle graph
    for node in lc_graph:
        lc = lc_graph.nodes[node]["cycle"]

        lc_graph.nodes[node]["start_frame"] = tr_graph.nodes[lc[0]]["frame"]
        lc_graph.nodes[node]["end_frame"] = tr_graph.nodes[lc[-1]]["frame"]

    return lc_graph


def delete_nodes(graph: nx.DiGraph, nodes_to_delete: list) -> nx.DiGraph:
    """Delete nodes while maintaining the connectivity

    Args:
        graph (nx.DiGraph): _description_
        nodes_to_delete (list): _description_

    Returns:
        nx.DiGraph: _description_
    """
    for node in nodes_to_delete:
        preds = list(graph.predecessors(node))
        succs = list(graph.successors(node))

        for p, s in product(preds, succs):
            graph.add_edge(p, s)

        graph.remove_node(node)

    return graph


def subsample_lineage(lineage: nx.DiGraph, subsampling_factor: int) -> nx.DiGraph:
    """Subsample lineage by only takeing nodes in every n-th (subsampling_factor) frame. Connectivity is maintained.

    Args:
        lineage (nx.DiGraph): lineage graph (needs the frame attributes)
        subsampling_factor (int): n-th frame will taken into account

    Returns:
        nx.DiGraph: Returns the pruned lineage only containing nodes of every n-th frame
    """

    # copy the lineage
    lineage = lineage.copy()

    # get all the frames in the lineage
    frames = list(sorted(np.unique([lineage.nodes[n]["frame"] for n in lineage.nodes])))

    # compute what frames to keep (every n-th)
    keep_frames = set(frames[::subsampling_factor])

    # create a list of nodes that are not inside the selected frames
    del_nodes = [
        n for n in lineage.nodes if lineage.nodes[n]["frame"] not in keep_frames
    ]

    # delete nodes (maintain connectivity)
    new_lineage = delete_nodes(lineage, del_nodes)

    # return the new lineage
    return new_lineage


def tracklet_to_tracking(ov: Overlay, tracklet_graph: nx.DiGraph) -> nx.DiGraph:
    """Compute a tracking graph based on the tracklet graph and an overlay

    Args:
        ov (Overlay): the current overlay
        tracklet_graph (nx.DiGraph): the corresponding tracklet_graph

    Returns:
        nx.DiGraph: the resulting tracking graph
    """
    tracking_graph = nx.DiGraph()

    label_lookup = {}

    for cont in ov:
        if cont.label in tracklet_graph.nodes:
            tracking_graph.add_node(cont.id, frame=cont.frame)

            label_lookup[cont.label] = label_lookup.get(cont.label, []) + [cont]

    for label in tracklet_graph.nodes:

        # get all the contours with this label
        contours = sorted(label_lookup[label], key=lambda c: c.frame)

        # add them sequentially
        for a, b in zip(contours, contours[1:]):
            tracking_graph.add_edge(a.id, b.id)

    for label in tracklet_graph.nodes:
        for succ in tracklet_graph.successors(label):
            # make the division edges
            tracking_graph.add_edge(
                label_lookup[label][-1].id, label_lookup[succ][0].id
            )

    return tracking_graph


def merge_incosistent_segmentation(
    sub_ov: Overlay, tracklet_graph: nx.DiGraph, num_timesteps=3
) -> tuple[Overlay, nx.DiGraph]:
    """Merges incosistent segmentation using the tracking information

    Args:
        sub_ov (Overlay): overlay containing segmentation information
        tracklet_graph (nx.DiGraph): the tracklet graph
        num_timesteps (int, optional): minimal number of timesteps that a segmentation must exist. Defaults to 3.

    Returns:
        tuple[Overlay, nx.DiGraph]: new overlay and tracklet_graph
    """

    def cond(n, graph, num_nodes=num_timesteps):
        """returns true if this is an event where the tracking is inconsistent and should be joined"""
        if graph.out_degree(n) != 2:
            return False

        children = sorted(graph.successors(n), key=graph.out_degree)

        # check whether we have a dead end and a continous cell
        if not (
            graph.out_degree(children[0]) == 0 and graph.out_degree(children[1]) >= 1
        ):
            return False

        # check that the dead end durtion is not too long
        dur = (
            graph.nodes[children[0]]["end_frame"]
            - graph.nodes[children[0]]["start_frame"]
        )
        if dur > num_nodes:
            return False

        return True

    # collect all the siblsings that should be joined
    to_join = []
    for n in tracklet_graph.nodes:
        # check the join condition
        if cond(n, tracklet_graph):

            children = sorted(
                tracklet_graph.successors(n), key=tracklet_graph.out_degree
            )
            to_join.append(children)

    new_ov = Overlay([])

    remove_labels = {join_set[0] for join_set in to_join}

    # create the new overay where masks are joined
    for i, ov in enumerate(sub_ov.timeIterator()):
        frame_label_set = {it.label for it in ov}

        to_add = []
        to_remove = []

        for join_set in to_join:
            if set(join_set).issubset(frame_label_set):
                # print(f"Frame: {i} -> Need to change overlay")

                def label_lookup(label, ov):
                    return [cont for cont in ov if cont.label == label][0]

                # print(ov.cont_lookup)
                polys = [
                    label_lookup(join_set[0], ov).polygon.buffer(2),
                    label_lookup(join_set[1], ov).polygon.buffer(5),
                ]

                res_poly = unary_union(polys)
                res_poly = res_poly.buffer(-5)

                if isinstance(res_poly, MultiPolygon):
                    area_before = res_poly.area
                    max_size_index = np.argmax([g.area for g in res_poly.geoms])
                    res_poly = res_poly.geoms[max_size_index]
                    logging.warning(
                        "Need to fix multipolygon. Area from %.2f to %.2f",
                        area_before,
                        res_poly.area,
                    )

                # this polygon needs to be added
                cont = Contour(
                    np.stack(res_poly.exterior.xy, axis=-1),
                    -1,
                    frame=i,
                    id=label_lookup(join_set[1], ov).id,
                    label=join_set[1],
                )
                # print(cont.coordinates)
                to_add.append(cont)
                to_remove.append(join_set[1])

        all_remove = remove_labels.union(set(to_remove))
        new_ov.add_contours(
            [cont for cont in ov if cont.label not in all_remove] + to_add
        )

    # remove joined labels from the tracklet graph
    remove_labels = {join_set[0] for join_set in to_join}
    for n in remove_labels:
        tracklet_graph.remove_node(n)

    # join tracklets (we have remove wrong divisions but still need to join the tracklets)
    tracklets_to_join = []

    for n in list(nx.dfs_preorder_nodes(tracklet_graph)):
        if tracklet_graph.out_degree(n) == 1:
            tracklets_to_join.append((n, list(tracklet_graph.successors(n))[0]))

    relabel_actions = {n: n for n in tracklet_graph.nodes}

    for a, b in tracklets_to_join:

        relabel_actions[b] = relabel_actions[a]

    # actually join the tracklets
    for b, a in relabel_actions.items():

        # join the two
        b_children = tracklet_graph.successors(b)

        # ensure connectivity
        for b_child in b_children:
            tracklet_graph.add_edge(a, b_child)

        # update end frame
        tracklet_graph.nodes[a]["end_frame"] = np.max(
            [tracklet_graph.nodes[a]["end_frame"], tracklet_graph.nodes[b]["end_frame"]]
        )

    # remove nodes
    tracklet_graph.remove_nodes_from(
        set(tracklet_graph.nodes).difference(relabel_actions.values())
    )

    for cont in new_ov:
        if isinstance(cont, Instance) and cont.label != relabel_actions[cont.label]:
            cont.mask = (cont.mask == cont.label) * relabel_actions[cont.label]
        cont.label = relabel_actions[cont.label]

    return new_ov, tracklet_graph



import os
import numpy as np
from PIL import ImageDraw, Image
import matplotlib.pyplot as plt

class_id_to_name_mapping = ["sports ball", "post"]


def plot_bounding_box(image, annotation_list):
    annotations = np.array(annotation_list)
    w, h = image.size

    plotted_image = ImageDraw.Draw(image)

    transformed_annotations = np.copy(annotations)
    transformed_annotations[:, [1, 3]] = annotations[:, [1, 3]] * w
    transformed_annotations[:, [2, 4]] = annotations[:, [2, 4]] * h

    transformed_annotations[:, 1] = transformed_annotations[:, 1] - (transformed_annotations[:, 3] / 2)
    transformed_annotations[:, 2] = transformed_annotations[:, 2] - (transformed_annotations[:, 4] / 2)
    transformed_annotations[:, 3] = transformed_annotations[:, 1] + transformed_annotations[:, 3]
    transformed_annotations[:, 4] = transformed_annotations[:, 2] + transformed_annotations[:, 4]

    for ann in transformed_annotations:
        obj_cls, x0, y0, x1, y1 = ann
        plotted_image.rectangle(((x0, y0), (x1, y1)))

        plotted_image.text((x0, y0 - 10), class_id_to_name_mapping[(int(obj_cls))])

    plt.imshow(np.array(image))
    plt.show()


# Get any random annotation file
annotation_file = "/home/johannes/projects/work/acia/yolo_dataset/annotations/train/img_000_00017.txt"
with open(annotation_file, "r") as file:
    annotation_list = file.read().split("\n")[:-1]
    annotation_list = [x.split(" ") for x in annotation_list]
    annotation_list = [[float(y) for y in x] for x in annotation_list]

# Get the corresponding image file
image_file = annotation_file.replace("annotations", "images").replace("txt", "png")
assert os.path.exists(image_file)

# Load the image
image = Image.open(image_file)

# Plot the Bounding Box
plot_bounding_box(image, annotation_list)



"""Module for Contour Proposal Networks"""

import celldetection as cd
import numpy as np
import torch
from tqdm.auto import tqdm

from acia.attribute import attribute_segmentation
from acia.base import Contour, Overlay


class CPNSegmenter:
    """Contour Proposal Networks segmenter: https://github.com/FZJ-INM1-BDA/celldetection"""

    def __init__(self, nms_thresh=0.4):

        # Load pretrained model
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.model = cd.fetch_model(
            "ginoro_CpnResNeXt101UNet-fbe875f1a3e5ce2c", check_hash=True
        ).to(self.device)
        self.model.nms_thresh = nms_thresh
        self.model.eval()

    def __call__(self, image_sequence):

        contours = []
        max_frame = 0
        for frame_id, img in enumerate(
            tqdm(image_sequence, desc="Perform segmentation...")
        ):
            # Load input
            img = img.raw
            print(img.dtype, img.shape, (img.min(), img.max()))

            if len(img.shape) == 3:
                # we have HxWxC
                # strip of last channel to make it grayscale
                img = img[..., 0]

            # convert to rgb
            img = np.stack((img,) * 3, axis=-1)

            # Run model
            with torch.no_grad():
                x = cd.to_tensor(
                    img, transpose=True, device=self.device, dtype=torch.float32
                )
                x = x / x.max()  # ensure 0..1 range
                x = x[
                    None
                ]  # add batch dimension: Tensor[3, h, w] -> Tensor[1, 3, h, w]
                y = self.model(x)

            frame_ov = y["contours"][0]
            torch_frame_ov = frame_ov.cpu().numpy()
            for cont in torch_frame_ov:
                contours.append(Contour(cont, -1, frame_id, 0))

            max_frame = frame_id

        overlay = Overlay(contours, frames=list(range(max_frame + 1)))

        attribute_segmentation(overlay, self)

        return overlay



"""Module for general visualization functionality
"""

from __future__ import annotations

import logging
from collections import deque
from datetime import timedelta
from pathlib import Path

import cv2
import matplotlib.pyplot as plt
import moviepy.editor as mpy
import networkx as nx
import numpy as np
import pint
import plotly.graph_objects as go
from matplotlib import font_manager
from PIL import Image, ImageDraw, ImageFont
from tqdm.auto import tqdm

from acia import ureg
from acia.base import BaseImage, ImageSequenceSource, Overlay
from acia.segm.local import InMemorySequenceSource, LocalImage, THWCSequenceSource

from .utils import strfdelta

# loda the deja vu sans default font
default_font = font_manager.findfont("DejaVu Sans")


def draw_scale_bar(
    image_iterator,
    xy_position: tuple[int, int],
    size_of_pixel,
    bar_width,
    bar_height,
    color=(255, 255, 255),
    font_size=25,
    font_path=default_font,
    background_color=None,
    background_margin_pixel=3,
):
    """Draws a scale bar on all images of an image sequence or iterable image array

    Args:
        image_iterator: image sequence or iterator over images
        xy_position (tuple[int, int]): lower left xy position of the scale bar
        size_of_pixel (_type_): metric size of a pixel (e.g. 0.007 * ureg.micrometer)
        bar_width (_type_): width of the scale bar (e.g. 5 * ureg.micrometer)
        short_title (str, optional): Short title of the unit to be displayed. Defaults to "μm".
        color (tuple, optional): Color of scale bar and text. Defaults to (255, 255, 255).
        font_size (int, optional): text font size. Defaults to 25.
        font_path (str, optional): text font. Defaults to "/usr/share/fonts/truetype/dejavu/DejaVuSerif.ttf".
        background_color: color for a potential background rectangle (e.g. (0, 0, 0)). Defaults to None (no background drawn).
        background_margin_pixel: pixels of margin for the background rectangle

    Yields:
        np.ndarray | LocalImage: Image in numpy format or LocalImage (depending on the input format)
    """

    # create pint quantities (values and units)
    bar_width = ureg.Quantity(bar_width)
    bar_height = ureg.Quantity(bar_height)
    size_of_pixel = ureg.Quantity(size_of_pixel)

    # load font
    font = ImageFont.truetype(font_path, font_size)

    # compute width and height of the scale bar in pixels (we need to round here)
    bar_pixel_width = int(
        np.round((bar_width / size_of_pixel).to_base_units().magnitude)
    )
    bar_pixel_height = int(
        np.round((bar_height / size_of_pixel).to_base_units().magnitude)
    )

    # extract position
    xstart, ystart = xy_position

    for image in image_iterator:

        # do we have a wrapped image?
        is_wrapped = isinstance(image, BaseImage)

        # unwrap if necessary
        if is_wrapped:
            image = image.raw

        # compute text size
        text = f"{bar_width:~P}"
        img_pil = Image.fromarray(image)
        draw = ImageDraw.Draw(img_pil)

        # get size of text
        left, top, right, bottom = draw.textbbox((xstart, ystart), text, font=font)

        text_width = right - left
        text_height = bottom - top

        if background_color:
            cv2.rectangle(
                image,
                (xstart - background_margin_pixel, ystart + background_margin_pixel),
                (
                    xstart + bar_pixel_width + background_margin_pixel,
                    ystart
                    - text_height
                    - bar_pixel_height
                    - 5
                    - background_margin_pixel,
                ),
                background_color,
                -1,
            )

        # draw scale bar
        cv2.rectangle(
            image,
            (xstart, ystart),
            (xstart + bar_pixel_width, ystart - bar_pixel_height),
            (255, 255, 255),
            -1,
        )

        img_pil = Image.fromarray(image)
        draw = ImageDraw.Draw(img_pil)

        # draw text centered and with distance to the scale bar
        draw.text(
            (
                xstart + bar_pixel_width / 2 - text_width / 2,
                ystart - text_height - bar_pixel_height - 10,
            ),
            text,
            fill=color,
            font=font,
        )

        # convert PIL image back to numpy
        image = np.array(img_pil)

        # do the image wrapping
        if is_wrapped:
            yield LocalImage(image)
        else:
            yield image


def draw_time(
    image_iterator,
    xy_position,
    time_step,
    color=(255, 255, 255),
    font_size=25,
    font_path=default_font,
    background_color=None,
    background_margin_pixel=3,
):
    """Draw time onto images

    Args:
        image_iterator (_type_): image sequence or iterator over images
        xy_position (tuple[int, int]): lower left xy position of the time text
        time_step (_type_): time step between images (e.g. 15 * ureg.minute or "15 minute")
        color (_type): Color of the time text. Defaults to (255, 255, 255) which is white.
        font_size (int, optional): text font size. Defaults to 25.
        font_path (str, optional): text font. Defaults to "/usr/share/fonts/truetype/dejavu/DejaVuSerif.ttf".
        background_color: color for a potential background rectangle (e.g. (0, 0, 0)). Defaults to None (no background drawn).
        background_margin_pixel: pixels of margin for the background rectangle

    Yields:
        _type_: _description_
    """

    time_step = ureg.Quantity(time_step)

    # load font
    font = ImageFont.truetype(font_path, font_size)

    for frame, image in enumerate(image_iterator):

        # do we have a wrapped image?
        is_wrapped = isinstance(image, BaseImage)

        # unwrap if necessary
        if is_wrapped:
            image = image.raw

        # convert to pillow image
        pil_image = Image.fromarray(image)
        draw = ImageDraw.Draw(pil_image)

        # extract time in hours and minutes
        time = (frame * time_step).to(ureg.hour)
        hours = int(np.floor(time.magnitude))
        minutes = int(np.round((time - hours * ureg.hour).to("minute").magnitude))

        time_text = f"Time: {hours:2d}:{minutes:02d} h"

        if background_color:
            # get size of text
            left, top, right, bottom = draw.textbbox(xy_position, time_text, font=font)

            text_width = right - left
            text_height = bottom - top

            x, y = xy_position

            cv2.rectangle(
                image,
                (x - background_margin_pixel, y - background_margin_pixel),
                (
                    x + text_width + background_margin_pixel,
                    y + text_height + background_margin_pixel + 5,
                ),
                background_color,
                -1,
            )

            # convert to pillow image
            pil_image = Image.fromarray(image)
            draw = ImageDraw.Draw(pil_image)

        # draw on image
        draw.text(xy_position, time_text, fill=color, font=font)

        # convert PIL image back to numpy
        image = np.array(pil_image)

        # do the image wrapping
        if is_wrapped:
            yield LocalImage(image)
        else:
            yield image


class VideoExporter:
    """
    Wrapper for opencv video writer. Simplifies usage
    """

    def __init__(self, filename, framerate, codec="MJPG"):
        self.filename = filename
        self.framerate = framerate
        self.out = None
        self.frame_height = None
        self.frame_width = None
        self.codec = codec

    def __del__(self):
        if self.out:
            self.close()

    def write(self, image):
        height, width = image.shape[:2]
        if self.out is None:
            self.frame_height, self.frame_width = image.shape[:2]
            self.out = cv2.VideoWriter(
                self.filename,
                cv2.VideoWriter_fourcc(*self.codec),
                self.framerate,
                (self.frame_width, self.frame_height),
            )
        if self.frame_height != height or self.frame_width != width:
            logging.warning(
                "You add images of different resolution to the VideoExporter. This may cause problems (e.g. black video output)!"
            )
        self.out.write(image)

    def close(self):
        if self.out:
            self.out.release()
            self.out = None

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if self.out is None:
            logging.warning(
                "Closing video writer without any images written and no video output generated! Did you forget to write the images="
            )
        self.close()


class VideoExporter2:
    """
    Wrapper for opencv video writer. Simplifies usage
    """

    def __init__(
        self, filename: Path, framerate: int, codec="mjpeg", ffmpeg_params=None
    ):
        self.filename = Path(filename)
        self.framerate = framerate
        self.codec = codec

        if ffmpeg_params is None:
            ffmpeg_params = []

        self.ffmpeg_params = ffmpeg_params

        self.images = []

    @staticmethod
    def default_vp9(
        filename: Path,
        framerate: int,
    ):
        ffmpeg_params = ["-crf", "30", "-b:v", "0", "-speed", "1"]
        return VideoExporter2(
            filename, framerate, codec="libvpx-vp9", ffmpeg_params=ffmpeg_params
        )

    @staticmethod
    def fast_vp9(
        filename: Path,
        framerate: int,
    ):
        ffmpeg_params = ["-crf", "35", "-b:v", "0", "-speed", "3"]
        return VideoExporter2(
            filename, framerate, codec="libvpx-vp9", ffmpeg_params=ffmpeg_params
        )

    @staticmethod
    def default_h264(
        filename: Path,
        framerate: int,
    ):
        ffmpeg_params = ["-crf", "30", "-preset", "fast"]
        return VideoExporter2(
            filename, framerate, codec="libx264", ffmpeg_params=ffmpeg_params
        )

    @staticmethod
    def default_h265(filename: Path, framerate: int):
        ffmpeg_params = ["-crf", "26", "-preset", "fast"]
        return VideoExporter2(
            filename, framerate, codec="libx265", ffmpeg_params=ffmpeg_params
        )

    @staticmethod
    def default_mjpg(filename: Path, framerate: int):
        ffmpeg_params = []
        return VideoExporter2(
            filename, framerate, codec="mjpeg", ffmpeg_params=ffmpeg_params
        )

    # av1 not yet supported
    #    @staticmethod
    #    def default_av1(filename: Path, framerate: int, ffmpeg_params=["-crf", "26", "-preset", "2", "-strict", "2"]):
    #        return VideoExporter2(filename, framerate, codec="libaom-av1", ffmpeg_params=ffmpeg_params)

    def write(self, image):
        self.images.append(image)

    def close(self):
        if len(self.images) == 0:
            logging.warning(
                "Closing video writer without any images written and no video output generated! Did you forget to write the images?"
            )
        else:
            # do the video rendering
            clip = mpy.ImageSequenceClip(
                list(self.images),
                fps=self.framerate,
            )
            clip.write_videofile(
                str(self.filename.absolute()),
                codec=self.codec,
                ffmpeg_params=self.ffmpeg_params,
                # verbose=False,
                # logger=None,
            )
            self.images = []

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()


def render_segmentation(
    imageSource: ImageSequenceSource,
    overlay: Overlay,
    cell_color=(255, 255, 0),
) -> ImageSequenceSource:
    """Render a video of the time-lapse including the segmentaiton information.

    Args:
        imageSource (ImageSequenceSource): Your time-lapse source object.
        Overlay ([type]): Your source of RoIs for the image (e.g. cells).
        cell_color: rgb color of the cell outlines
    """

    if overlay is None:
        # when we have no rois -> create iterator that always returns None
        def always_none():
            while True:
                yield None

        overlay = iter(always_none())

    images = []

    for image, frame_overlay in tqdm(
        zip(imageSource, overlay.timeIterator()), desc="Render cell segmentation..."
    ):
        # extract the numpy image
        if isinstance(image, BaseImage):
            image = image.raw
        elif isinstance(image, np.ndarray):
            pass
        else:
            raise Exception("Unsupported image type!")

        # copy image as we draw onto it
        image = np.copy(image)

        if len(image.shape) == 2:
            # convert to grayscale if needed
            image = np.stack((image,) * 3, axis=-1)

        if len(image.shape) != 3 or image.shape[2] != 3:
            logging.warning(
                "Your images are in the wrong shape! The shape of an image is %s but we need (height, width, 3)! This is likely to cause an error!",
                image.shape,
            )

        # Draw overlay
        if frame_overlay:
            image = frame_overlay.draw(image, cell_color)  # RGB format

        images.append(image)

    # return as sequence source again
    return InMemorySequenceSource(np.stack(images))


def render_cell_centers(
    image_source: ImageSequenceSource | np.ndarray,
    overlay: Overlay,
    center_color=(255, 255, 0),
    center_size=3,
) -> ImageSequenceSource:
    """Render a image sequence of the time-lapse with the cell centers.

    Args:
        imageSource (ImageSequenceSource): Your time-lapse source object.
        overlay (Overlay, optional): Your source of RoIs for the image (e.g. cells).
        center_color (tuple, optional): RGB color of the cell center circle. Defaults to (255, 255, 0).
        center_size (int, optional): Radius of the cell center circle (in pixels). Defaults to 3.

    Raises:
        ValueError: If we recognize unsupported image type or format

    Returns:
        ImageSequenceSource: The rendered image sequence
    """

    if overlay is None:
        # when we have no rois -> create iterator that always returns None
        def always_none():
            while True:
                yield None

        overlay = iter(always_none())

    images = []

    for image, frame_overlay in tqdm(
        zip(image_source, overlay.timeIterator()), desc="Render cell centers..."
    ):
        # extract the numpy image
        if isinstance(image, BaseImage):
            image = image.raw
        elif isinstance(image, np.ndarray):
            pass
        else:
            raise ValueError("Unsupported image type!")

        # copy image as we draw onto it
        image = np.copy(image)

        # Draw overlay
        if frame_overlay:

            # compute all centers
            centers = [cont.center for cont in frame_overlay]

            for center in centers:
                int_center = tuple(map(int, center))

                cv2.circle(image, int_center, center_size, center_color, -1)

        images.append(image)

    image_stack = np.stack(images)

    if isinstance(ImageSequenceSource, np.ndarray):
        # return as raw numpy stack
        return image_stack
    else:
        # return as sequence source again
        return InMemorySequenceSource(image_stack)


def render_tracking(
    image_source: ImageSequenceSource,
    overlay: Overlay,
    tracking_graph: nx.DiGraph,
) -> ImageSequenceSource:
    """Render the tracking to an image source

    Args:
        image_source (ImageSequenceSource): Image source
        overlay (Overlay): overla of cell detections (for center points)
        tracking_graph (nx.DiGraph): the tracking graph where every cell detection is a node in the graph.

    Returns:
        ImageSequenceSource: Rendered image source
    """

    images = []

    contour_lookup = {cont.id: cont for cont in overlay}

    for image, frame_overlay in zip(
        tqdm(image_source, desc="Render cell tracking paths..."), overlay.timeIterator()
    ):

        np_image = np.copy(image.raw)

        if len(np_image.shape) == 2:
            # convert to grayscale if needed
            np_image = np.stack((np_image,) * 3, axis=-1)

        if len(np_image.shape) != 3 or np_image.shape[2] != 3:
            logging.warning(
                "Your images are in the wrong shape! The shape of an image is %s but we need (height, width, 3)! This is likely to cause an error!",
                image.shape,
            )

        for cont in frame_overlay:
            if cont.id in tracking_graph.nodes:
                edges = tracking_graph.out_edges(cont.id)

                born = tracking_graph.in_degree(cont.id) == 0

                for edge in edges:
                    source = contour_lookup[edge[0]].center
                    target = contour_lookup[edge[1]].center

                    line_color = (255, 0, 0)  # rgb: red

                    if len(edges) > 1:
                        line_color = (0, 0, 255)  # bgr: blue

                    cv2.line(
                        np_image,
                        tuple(map(int, source)),
                        tuple(map(int, target)),
                        line_color,
                        thickness=3,
                    )

                    if born:
                        cv2.circle(
                            np_image,
                            tuple(map(int, source)),
                            3,
                            (203, 192, 255),
                            thickness=1,
                        )

                if len(edges) == 0:
                    cv2.rectangle(
                        np_image,
                        np.array(cont.center).astype(np.int32) - 2,
                        np.array(cont.center).astype(np.int32) + 2,
                        (203, 192, 255),
                    )

        images.append(np_image)

    return InMemorySequenceSource(images)


def render_video(
    image_source: ImageSequenceSource,
    filename: str,
    framerate: int,
    codec: str,
    ffmpeg_params: list[str] = None,
) -> None:
    """Render video

    Args:
        image_source (ImageSequenceSource): sequence of images
        filename (str): video filename
        framerate (int): framerate of the video
        codec (str): the codec for video encoding
    """

    with VideoExporter2(
        str(filename), framerate=framerate, codec=codec, ffmpeg_params=ffmpeg_params
    ) as ve:
        for im in tqdm(image_source, desc="Encoding video..."):

            image = im.raw

            if len(image.shape) == 2:
                # convert to grayscale if needed
                image = np.stack((image,) * 3, axis=-1)

            if len(image.shape) != 3 or image.shape[2] != 3:
                logging.warning(
                    "Your images are in the wrong shape! The shape of an image is %s but we need (height, width, 3)! This is likely to cause an error!",
                    image.shape,
                )

            ve.write(image)


def render_scalebar(
    image_source: Overlay,
    xy_position: tuple[int | float, int | float],
    size_of_pixel: pint.Quantity,
    bar_width: pint.Quantity,
    bar_height: pint.Quantity,
    color=(255, 255, 255),
    font_size=25,
    font_path=default_font,
    background_color: tuple[int, int, int] = None,
    background_margin_pixel=3,
    show_text=True,
) -> ImageSequenceSource:
    """Draws a scale bar on all images of an image sequence or iterable image array

    Args:
        image_source (Overlay): image sequence or iterator over images
        xy_position (tuple[int, int]): lower left xy position of the scale bar
        size_of_pixel (pint.Quantity): metric size of a pixel (e.g. 0.007 * ureg.micrometer)
        bar_width (pint.Quantity): width of the scalebar (e.g. 5 * ureg.micrometer). Also the text over the bar.
        bar_height (pint.Quantity): height of the scalebar.
        color (tuple, optional): Color of the scalebar and text. Defaults to (255, 255, 255).
        font_size (int, optional): font size of the text. Defaults to 25.
        font_path (_type_, optional): path to the font. Defaults to default_font.
        background_color (tuple[int, int, int], optional): Color of the background. None draws no background. Defaults to None.
        background_margin_pixel (int, optional): Margin of the background box. Defaults to 3.
        show_text (bool, optional): If true shows the bar width as text above the bar. Defaults to True.

    Returns:
        ImageSequenceSource: Rendered image sequence
    """

    # create pint quantities (values and units)
    bar_width = ureg.Quantity(bar_width)
    bar_height = ureg.Quantity(bar_height)
    size_of_pixel = ureg.Quantity(size_of_pixel)

    # load font
    font = ImageFont.truetype(font_path, font_size)

    # compute width and height of the scale bar in pixels (we need to round here)
    bar_pixel_width = int(
        np.round((bar_width / size_of_pixel).to_base_units().magnitude)
    )
    bar_pixel_height = int(
        np.round((bar_height / size_of_pixel).to_base_units().magnitude)
    )

    image_height, image_width = image_source.get_frame(0).raw.shape[:2]

    # extract position
    xstart, ystart = xy_position

    # Allow relative positioning
    if isinstance(xstart, float):
        if xstart > 1.0:
            raise ValueError(
                f"If using float (x,y) position coordinates they have to be below 1. Your x position is {xstart}"
            )
        xstart = int(np.round(image_width * xstart))

    if isinstance(ystart, float):
        if ystart > 1.0:
            raise ValueError(
                f"If using float (x,y) position coordinates they have to be below 1. Your x position is {xstart}"
            )
        ystart = int(np.round(image_height * ystart))

    images = []

    for image in tqdm(image_source, desc="Render scale bar..."):

        # do we have a wrapped image?
        is_wrapped = isinstance(image, BaseImage)

        # unwrap if necessary
        if is_wrapped:
            image = image.raw

        image = np.copy(image)

        # compute text size
        text = f"{bar_width:~P}"
        img_pil = Image.fromarray(image)
        draw = ImageDraw.Draw(img_pil)

        # get size of text
        left, top, right, bottom = draw.textbbox((xstart, ystart), text, font=font)

        text_width = right - left
        text_height = bottom - top

        if background_color:
            cv2.rectangle(
                image,
                (xstart - background_margin_pixel, ystart + background_margin_pixel),
                (
                    xstart + bar_pixel_width + background_margin_pixel,
                    ystart
                    - text_height
                    - bar_pixel_height
                    - 5
                    - background_margin_pixel,
                ),
                background_color,
                -1,
            )

        # draw scale bar
        cv2.rectangle(
            image,
            (xstart, ystart),
            (xstart + bar_pixel_width, ystart - bar_pixel_height),
            (255, 255, 255),
            -1,
        )

        if show_text:
            img_pil = Image.fromarray(image)
            draw = ImageDraw.Draw(img_pil)

            # draw text centered and with distance to the scale bar
            draw.text(
                (
                    xstart + bar_pixel_width / 2 - text_width / 2,
                    ystart - text_height - bar_pixel_height - 10,
                ),
                text,
                fill=color,
                font=font,
            )

            # convert PIL image back to numpy
            image = np.array(img_pil)

        images.append(image)

    # combine all images
    image_stack = np.stack(images)

    if isinstance(ImageSequenceSource, np.ndarray):
        # return as raw numpy stack
        return image_stack
    else:
        # return as sequence source again
        return InMemorySequenceSource(image_stack)


def render_time(
    image_source: ImageSequenceSource,
    xy_position: tuple[int | float, int | float],
    timepoints: list[pint.Quantity | timedelta],
    time_format="{H:02}h {M:02}m",
    color=(255, 255, 255),
    font_size=25,
    font_path=default_font,
    background_color: tuple[int, int, int] = None,
    background_margin_pixel=3,
) -> ImageSequenceSource:
    """Draw time onto images

    Args:
        image_source (ImageSequenceSource): image sequence of the time-lapse
        xy_position (tuple[int]): lower left xy position of the formatted time text
        timepoints (list[pint.Quantity  |  timedelta]): timepoints of the individual frames
        time_format (str, optional): Timeformat for rendering the time to the images. Defaults to "{H:02}h {M:02}m".
        color (tuple, optional): Color of the time text. Defaults to (255, 255, 255).
        font_size (int, optional): Fontsize of the time text. Defaults to 25.
        font_path (_type_, optional): Path to the rendering font. Defaults to default_font.
        background_color (tuple[int, int, int], optional): Color of the background box. None does not draw any background box. Defaults to None.
        background_margin_pixel (int, optional): Margin of the background box. Defaults to 3.

    Returns:
        ImageSequenceSource: Rendered image sequence
    """

    # load font
    font = ImageFont.truetype(font_path, font_size)

    images = []

    image_height, image_width = image_source.get_frame(0).raw.shape[:2]

    # extract position
    xstart, ystart = xy_position

    # Allow relative positioning
    if isinstance(xstart, float):
        if xstart > 1.0:
            raise ValueError(
                f"If using float (x,y) position coordinates they have to be below 1. Your x position is {xstart}"
            )
        xstart = int(np.round(image_width * xstart))

    if isinstance(ystart, float):
        if ystart > 1.0:
            raise ValueError(
                f"If using float (x,y) position coordinates they have to be below 1. Your x position is {xstart}"
            )
        ystart = int(np.round(image_height * ystart))

    for image, timepoint in zip(tqdm(image_source, desc="Render time..."), timepoints):

        if isinstance(timepoint, pint.Quantity):
            timepoint = timedelta(seconds=float(timepoint.to(ureg.seconds).magnitude))

        # do we have a wrapped image?
        is_wrapped = isinstance(image, BaseImage)

        # unwrap if necessary
        if is_wrapped:
            image = image.raw

        image = np.copy(image)

        # convert to pillow image
        pil_image = Image.fromarray(image)
        draw = ImageDraw.Draw(pil_image)

        time_text = strfdelta(timepoint, fmt=time_format)

        if background_color:
            # get size of text
            left, top, right, bottom = draw.textbbox(xy_position, time_text, font=font)

            text_width = right - left
            text_height = bottom - top

            x, y = (xstart, ystart)

            cv2.rectangle(
                image,
                (x - background_margin_pixel, y - background_margin_pixel),
                (
                    x + text_width + background_margin_pixel,
                    y + text_height + background_margin_pixel + 5,
                ),
                background_color,
                -1,
            )

            # convert to pillow image
            pil_image = Image.fromarray(image)
            draw = ImageDraw.Draw(pil_image)

        # draw on image
        draw.text(xy_position, time_text, fill=color, font=font)

        # convert PIL image back to numpy
        image = np.array(pil_image)

        images.append(image)

    # combine all images
    image_stack = np.stack(images)

    if isinstance(ImageSequenceSource, np.ndarray):
        # return as raw numpy stack
        return image_stack
    else:
        # return as sequence source again
        return InMemorySequenceSource(image_stack)


def colorize_instance_mask(
    instance_mask, background_color=(0, 0, 0), seed=42, color_lut=None
) -> np.ndarray:
    """
    Convert instance mask to an RGB image with random colors per instance (no loop).

    Parameters:
        instance_mask (np.ndarray): 2D array of shape (H, W) with integer instance IDs.
        background_color (tuple): RGB color for background (default black).
        seed (int): Random seed for consistent coloring.
        color_lut (np.ndarray): Ix3 lookup map for instance colors (I)

    Returns:
        np.ndarray: Colored mask of shape (H, W, 3), dtype=uint8.
    """
    unique_ids = np.unique(instance_mask)
    unique_ids = unique_ids[unique_ids != 0]  # Exclude background (assumed to be 0)

    if len(unique_ids) == 0:
        return np.zeros((*instance_mask.shape, 3), dtype=np.uint8)

    # Map instance IDs to color lookup table (LUT)
    rng = np.random.default_rng(seed)
    if color_lut is None:
        color_lut = np.zeros((np.max(unique_ids) + 1, 3), dtype=np.uint8)
        # color_lut[unique_ids] = rng.integers(0, 256, size=(len(unique_ids), 3), dtype=np.uint8)
        color_lut = rng.integers(
            0, 256, size=(np.max(unique_ids) + 1, 3), dtype=np.uint8
        )
        color_lut[0] = background_color

    # Map colors to mask using LUT
    colored_mask = color_lut[instance_mask]

    return colored_mask


def render_segmentation_mask(
    source: ImageSequenceSource, overlay: Overlay, alpha=0.8
) -> THWCSequenceSource:
    """Render cell segmentation based on masks with random colors

    Args:
        source (ImageSequenceSource): the time-lapse sequence source
        overlay (Overlay): the corresponding overlay. WARNING: all instances need to be based on masks!
        alpha (float, optional): The opacity of the masked image. Defaults to 0.8.

    Returns:
        THWCSequenceSource: TxHxWx3 sequence
    """
    return_images = []

    for im, ov in zip(
        tqdm(source, desc="Render segmentation masks..."), overlay.time_iterator()
    ):
        im = np.copy(im.raw)

        colored_mask = np.zeros_like(im)

        for cont in ov:
            # render the masks based on the first contour mask in the frame
            colored_mask = colorize_instance_mask(cont.mask)
            break

        # Alpha blend with original image
        overlay = cv2.addWeighted(
            im.astype(np.float32), alpha, colored_mask.astype(np.float32), 1 - alpha, 0
        ).astype(np.uint8)

        # use the original image where no overlay is availabel
        binary_mask = np.stack((np.max(colored_mask, axis=-1),) * 3, axis=-1)
        overlay = np.where(binary_mask, overlay, im)

        return_images.append(overlay)

    # return the new time-lapse
    return THWCSequenceSource(np.stack(return_images, axis=0))


def render_tracking_mask(
    source: ImageSequenceSource,
    overlay: Overlay,
    alpha=0.8,
    show_label_numbers=False,
    seed=42,
) -> THWCSequenceSource:
    """Render tracking and use the label colors for the masks

    Args:
        source (ImageSequenceSource): the time-lapse sequence source
        overlay (Overlay): the corresponding overlay. WARNING: all instances need to be based on masks!
        alpha (float, optional): The opacity of the masked image. Defaults to 0.8.

    Returns:
        THWCSequenceSource: TxHxWx3 sequence
    """
    return_images = []

    # generate color LUT (persistent for labels)
    rng = np.random.default_rng(seed)
    unique_labels = np.unique([0] + [cont.label for cont in overlay])
    color_lut = rng.integers(
        0, 256, size=(np.max(unique_labels) + 1, 3), dtype=np.uint8
    )
    color_lut[0] = (0, 0, 0)

    for im, ov in zip(
        tqdm(source, desc="Render tracking mask..."), overlay.time_iterator()
    ):
        im = np.copy(im.raw)

        h, w = im.shape[:2]

        label_mask = np.zeros((h, w), dtype=np.uint32)

        for cont in ov:
            # print(f"Label: {cont.label}")
            cell_mask = (cont.binary_mask * cont.label).astype(np.uint16)
            label_mask = np.maximum(label_mask, cell_mask)

            # render label numbers if necessary
            if show_label_numbers:
                cv2.putText(
                    im,
                    f"{cont.label}",
                    np.array(cont.center).astype(int),
                    cv2.FONT_HERSHEY_SIMPLEX,
                    0.5,
                    (255, 0, 0),
                    1,
                    cv2.LINE_AA,
                )

        # render the masks based on the labels
        colored_mask = colorize_instance_mask(label_mask, color_lut=color_lut)

        # Alpha blend with original image
        overlay = cv2.addWeighted(
            im.astype(np.float32), alpha, colored_mask.astype(np.float32), 1 - alpha, 0
        ).astype(np.uint8)

        # use the original image where no overlay is availabel
        binary_mask = np.stack((np.max(colored_mask, axis=-1),) * 3, axis=-1)
        overlay = np.where(binary_mask, overlay, im)

        return_images.append(overlay)

    # return the new time-lapse
    return THWCSequenceSource(np.stack(return_images, axis=0))


def hierarchy_pos_loop_multi(
    G, roots, width=1.0, vert_gap=0.25, vert_loc=0, xcenter=0.5, sep=0.1
):
    """
    Computes 2D positions for nodes in a forest (multiple-rooted trees) for visualization.

    Parameters
    ----------
    G : networkx.DiGraph
        The directed graph representing the forest or collection of trees.
    roots : list
        List of root node labels (each representing a separate tree).
    width : float
        Total horizontal span to use for plotting all trees.
    vert_gap : float
        Vertical gap between generations (levels).
    vert_loc : float
        Y-coordinate for the root nodes (typically 0).
    xcenter : float
        X-coordinate of the overall forest center.
    sep : float
        Extra horizontal gap between adjacent trees (as a fraction of total width).

    Returns
    -------
    pos : dict
        A dictionary mapping node labels to (x, y) positions for plotting.
    """

    def count_leaves(node):
        """
        Count the number of leaf nodes (tips) under a given node.
        Used to space subtrees proportionally.
        """
        queue = deque([node])
        leaves = 0
        while queue:
            curr = queue.popleft()
            children = list(G.neighbors(curr))
            if not children:
                leaves += 1
            else:
                queue.extend(children)
        return leaves

    n_roots = len(roots)
    leaf_counts = [count_leaves(root) for root in roots]
    total_leaves = sum(leaf_counts)
    forest_width = width - sep * (n_roots - 1)  # width minus all inter-tree gaps
    pos = {}
    x_left = xcenter - width / 2  # far left of forest

    # Layout each tree side by side
    for _, (root, leaves) in enumerate(zip(roots, leaf_counts)):
        tree_width = forest_width * (leaves / total_leaves)  # wider for bigger trees
        xcenter_tree = x_left + tree_width / 2
        node_queue = deque()
        node_queue.append((root, xcenter_tree, vert_loc, tree_width, 0))
        while node_queue:
            node, xcenter_here, y_here, width_here, depth = node_queue.popleft()
            children = list(G.neighbors(node))
            if not children:
                # It's a leaf/tip: assign position directly
                pos[node] = (xcenter_here, y_here)
            else:
                # Compute leaves in each child subtree to space children
                subtree_leaves = []
                for child in children:
                    n_leaves = count_leaves(child)
                    subtree_leaves.append(n_leaves)
                total = sum(subtree_leaves)
                x_left_child = xcenter_here - width_here / 2
                for j, child in enumerate(children):
                    w = width_here * (subtree_leaves[j] / total)
                    xc = x_left_child + w / 2
                    node_queue.append((child, xc, y_here - vert_gap, w, depth + 1))
                    x_left_child += w
                pos[node] = (xcenter_here, y_here)
        x_left += tree_width + sep  # move to the right for the next tree
    return pos


def subtree_from_roots(G, roots):
    """
    Returns a subgraph containing all nodes reachable from the given root nodes.
    """
    nodes = set()
    for root in roots:
        nodes.add(root)
        descendants = nx.descendants(G, root)
        nodes.update(descendants)
    return G.subgraph(nodes).copy()


def plot_lineage_tree(
    G,
    pos,
    mode="vertical",
    flip_horizontal=False,
    flip_vertical=False,
    tick_length=0.02,
    branch_color="navy",
    tick_color="black",
    figsize=(10, 6),
    draw_labels=False,
    label_fontsize=10,
    label_offset=0.01,
    y_attr=None,
    ax=None,
    label_attr=None,
):
    """
    Plot a lineage tree or forest with L-shaped (90°) branches, ticks at each node,
    and (optionally) node labels. Supports vertical/horizontal orientation and flipping.
    Can use a custom node property (y_attr) for vertical positioning (e.g., "time").

    Parameters
    ----------
    G : networkx.DiGraph
        The directed graph (single tree or forest).
    pos : dict
        Mapping from node to (x, y) coordinates (from a hierarchy layout).
    mode : str
        'vertical' (roots at top, downward) or 'horizontal' (roots at left, rightwards).
    flip_horizontal : bool
        If True, horizontal mode is flipped (roots at right, tree grows left).
    flip_vertical : bool
        If True, vertical mode is flipped (roots at bottom, tree grows upward).
    tick_length : float
        Length of the small tick at each node.
    branch_color : str
        Color of tree branches.
    tick_color : str
        Color of node ticks.
    figsize : tuple
        Matplotlib figure size.
    draw_labels : bool
        If True, draws node names at their positions.
    label_fontsize : int
        Font size for node labels.
    label_offset : float
        Offset for label position relative to node.
    y_attr : str or None
        If set, uses G.nodes[node][y_attr] for y coordinate of each node. Otherwise uses layout y.
    """
    if ax is None:
        plt.figure(figsize=figsize)
        ax = plt.gca()

    # If y_attr is set, override y in pos with the node's property value
    if y_attr:
        new_pos = {}
        for node, (x, _) in pos.items():
            y = G.nodes[node].get(y_attr, None)
            if y is not None:
                new_pos[node] = (x, y)
            else:
                new_pos[node] = (x, pos[node][1])  # fallback to computed y
        pos = new_pos

    x_vals = [x for x, y in pos.values()]
    y_vals = [y for x, y in pos.values()]
    x_max = max(x_vals) if x_vals else 0
    y_max = max(y_vals) if y_vals else 0

    for parent, child in G.subgraph(list(pos.keys())).edges():
        x0, y0 = pos[parent]
        x1, y1 = pos[child]
        # Optionally flip axes for various modes
        X0 = x_max - x0 if flip_horizontal else x0
        X1 = x_max - x1 if flip_horizontal else x1
        Y0 = y_max - y0 if flip_vertical else y0
        Y1 = y_max - y1 if flip_vertical else y1

        if mode == "vertical":
            # Classic tree: root at top (or bottom if flipped), L-branches down
            plt.plot([x0, x0], [Y0, Y1], color=branch_color, linewidth=2)
            plt.plot([x0, x1], [Y1, Y1], color=branch_color, linewidth=2)
        elif mode == "horizontal":
            # Horizontal: root at left (or right if flipped), L-branches across
            plt.plot([Y0, Y1], [X0, X0], color=branch_color, linewidth=2)
            plt.plot([Y1, Y1], [X0, X1], color=branch_color, linewidth=2)
        else:
            raise ValueError("mode must be 'vertical' or 'horizontal'")

    # Draw ticks and optional labels
    for node, (x, y) in pos.items():
        X = x_max - x if flip_horizontal else x
        Y = y_max - y if flip_vertical else y

        if label_attr is None:
            label = str(node)
        else:
            label = G.nodes[node][label_attr]

        if mode == "vertical":
            if tick_color is not None:
                plt.plot(
                    [x - tick_length / 2, x + tick_length / 2],
                    [Y, Y],
                    color=tick_color,
                    linewidth=2,
                )
            if draw_labels and G.out_degree(node) > 1:
                label_x = x + tick_length + label_offset
                ha = "left"
                plt.text(
                    label_x,
                    Y,
                    label,
                    fontsize=label_fontsize,
                    va="center",
                    ha=ha,
                    color="darkgreen",
                )
        elif mode == "horizontal":
            if tick_color is not None:
                plt.plot(
                    [y, y],
                    [X - tick_length / 2, X + tick_length / 2],
                    color=tick_color,
                    linewidth=2,
                )
            if draw_labels and G.out_degree(node) > 1:
                offset = label_offset if not flip_horizontal else -label_offset
                ha = "left" if not flip_horizontal else "right"
                plt.text(
                    y + offset,
                    X,
                    label,
                    fontsize=label_fontsize,
                    va="center",
                    ha=ha,
                    color="darkgreen",
                )
    # ax.axis('off')
    flip_info = []
    if flip_vertical:
        flip_info.append("vertically")
    if flip_horizontal:
        flip_info.append("horizontally")


###########################################
# Add new lineage rendering functionality #
###########################################


def compute_lineage_y(G, time_feature="t"):
    """
    Assign a y-position to each node using a tidy tree layout.

    Parameters
    ----------
    G : nx.DiGraph
        The lineage graph.
    time_feature : str
        The node attribute that encodes time.

    Returns
    -------
    assigned_y : dict
        Mapping from node to y-coordinate (float).
    """
    roots = [n for n in G.nodes if G.in_degree(n) == 0]
    assigned_y = {}
    next_y = [0]

    def assign_y_iterative():
        # Assign unique y-coordinates to all tips, then propagate up for inner nodes.
        stack = []
        visited = set()
        # Start with roots (nodes with no parents), sorted by time
        for root in sorted(roots, key=lambda n: G.nodes[n][time_feature]):
            stack.append((root, 0))
            while stack:
                node, depth = stack.pop()
                if node in visited:
                    continue
                children = list(G.successors(node))
                if not children:
                    # Assign new y to each leaf node
                    assigned_y[node] = next_y[0]
                    next_y[0] += 1
                else:
                    # Process children before parent (postorder)
                    stack.append((node, depth))
                    for child in reversed(children):
                        if child not in visited:
                            stack.append((child, depth + 1))
                    visited.add(node)
                    continue
                visited.add(node)
        # For internal nodes, set y as average of children
        for root in roots:
            postorder = list(nx.dfs_postorder_nodes(G, source=root))
            for node in postorder:
                children = list(G.successors(node))
                if children:
                    assigned_y[node] = sum(assigned_y[c] for c in children) / len(
                        children
                    )

    assign_y_iterative()
    return assigned_y


def extract_lineage_plotdata(
    G, assigned_y, time_feature="t", label_name=None, orientation="horizontal"
):
    """
    Collect all node and edge positions and hover info for plotting.

    Parameters
    ----------
    G : nx.DiGraph
        The lineage graph.
    assigned_y : dict
        Node-to-y mapping (from compute_lineage_y).
    time_feature : str
        The node attribute that encodes time.
    label_name : str or None
        Which node attribute to use for the label (default: node name).
    orientation : str
        'horizontal' or 'vertical' for layout.

    Returns
    -------
    data : dict
        Contains:
            - xs, ys: node x and y positions
            - node_ids: node names
            - node_labels: text for labels
            - hover_texts: HTML hover text for Plotly
            - edge_xs, edge_ys: positions for edges
            - births_x, births_y: positions for birth nodes
            - ends_x, ends_y: positions for end nodes
    """
    xs, ys, node_ids, node_labels, hover_texts = [], [], [], [], []
    for n in G.nodes:
        t = G.nodes[n][time_feature]
        y = assigned_y[n]
        xs.append(t if orientation == "horizontal" else y)
        ys.append(y if orientation == "horizontal" else t)
        node_ids.append(n)
        # Node label for drawing
        if label_name is None:
            label = str(n)
        else:
            label = str(G.nodes[n][label_name]) if label_name in G.nodes[n] else str(n)
        node_labels.append(label)
        # Build a pseudo-table hover (as HTML using <br> for newlines)
        features = G.nodes[n]
        if features:
            # For "aligned" look: pad keys to equal length
            maxk = max((len(str(k)) for k in features), default=1)
            fmt = lambda k, v, maxk: f"{str(k).ljust(maxk)} : {v}<br>"
            feat_lines = "".join(fmt(k, v, maxk) for k, v in features.items())
            hover_html = f"<b>Node:</b> {n}<br><span style='font-family:monospace'>{feat_lines}</span>"
        else:
            hover_html = f"<b>Node:</b> {n}"
        hover_texts.append(hover_html)

    # Edges: a list of (x0,x1), (y0,y1) for each edge
    edge_xs, edge_ys = [], []
    for n in G.nodes:
        t0 = G.nodes[n][time_feature]
        y0 = assigned_y[n]
        for c in G.successors(n):
            t1 = G.nodes[c][time_feature]
            y1 = assigned_y[c]
            if orientation == "horizontal":
                edge_xs.append([t0, t1])
                edge_ys.append([y0, y1])
            else:
                edge_xs.append([y0, y1])
                edge_ys.append([t0, t1])

    # Find birth and end nodes
    births_x, births_y, ends_x, ends_y = [], [], [], []
    for n in G.nodes:
        t = G.nodes[n][time_feature]
        y = assigned_y[n]
        if G.in_degree(n) == 0:
            # New/birth nodes (no parents)
            if orientation == "horizontal":
                births_x.append(t)
                births_y.append(y)
            else:
                births_x.append(y)
                births_y.append(t)
        if G.out_degree(n) == 0:
            # End nodes (no children)
            if orientation == "horizontal":
                ends_x.append(t)
                ends_y.append(y)
            else:
                ends_x.append(y)
                ends_y.append(t)
    return dict(
        xs=xs,
        ys=ys,
        node_ids=node_ids,
        node_labels=node_labels,
        hover_texts=hover_texts,
        edge_xs=edge_xs,
        edge_ys=edge_ys,
        births_x=births_x,
        births_y=births_y,
        ends_x=ends_x,
        ends_y=ends_y,
    )


def plot_cell_lineage(
    G,
    time_feature="t",
    orientation="horizontal",
    show_label=True,
    label_name=None,
    node_marker="o",
    node_ms=6,
    line_color="blue",
    line_lw=2,
    mark_births=False,
    birth_color="red",
    birth_marker=None,
    birth_ms=12,
    mark_ends=False,
    end_color="orange",
    end_marker="s",
    end_ms=10,
    ax=None,
    interactive_tooltip=False,
):
    """
    Draw a cell lineage tree as a static matplotlib plot.

    Parameters
    ----------
    G : nx.DiGraph
        The lineage graph.
    time_feature : str
        Node attribute for x-axis (typically time).
    orientation : str
        'horizontal' (time on x) or 'vertical' (time on y).
    show_label : bool
        Show node labels on the plot.
    label_name : str or None
        Node attribute to use for label (default: node name).
    node_marker : str
        Marker style for all nodes.
    node_ms : int
        Marker size for all nodes.
    line_color : str
        Color for edges and nodes.
    line_lw : int or float
        Edge line width.
    mark_births : bool
        Mark new tracks with a special marker/color.
    birth_color : str
        Color for birth marker.
    birth_marker : str
        Marker for birth nodes.
    birth_ms : int
        Size for birth marker.
    mark_ends : bool
        Mark track ends with a special marker/color.
    end_color : str
        Color for end marker.
    end_marker : str
        Marker for end nodes.
    end_ms : int
        Size for end marker.
    ax : plt.Axes or None
        If given, draw into this axes.
    interactive_tooltip : bool
        If True and mplcursors is installed, enables interactive node tooltips.
    """
    assigned_y = compute_lineage_y(G, time_feature)
    data = extract_lineage_plotdata(
        G, assigned_y, time_feature, label_name, orientation
    )

    if ax is None:
        fig, ax = plt.subplots(figsize=(12, 6))

    # Draw edges
    for x, y in zip(data["edge_xs"], data["edge_ys"]):
        ax.plot(x, y, "-", color=line_color, lw=line_lw)

    # Draw nodes
    main_nodes = ax.scatter(
        data["xs"], data["ys"], marker=node_marker, color=line_color, s=node_ms**2
    )
    if show_label:
        for x, y, label in zip(data["xs"], data["ys"], data["node_labels"]):
            ax.text(x, y + 0.12, label, fontsize=7, ha="center", va="bottom")

    # Draw markers for births and ends
    if mark_births and data["births_x"]:
        marker = (
            birth_marker
            if birth_marker
            else (">" if orientation == "horizontal" else "v")
        )
        ax.scatter(
            data["births_x"],
            data["births_y"],
            marker=marker,
            color=birth_color,
            s=birth_ms**2,
            zorder=5,
            alpha=0.9,
            edgecolor="k",
        )
    if mark_ends and data["ends_x"]:
        ax.scatter(
            data["ends_x"],
            data["ends_y"],
            marker=end_marker,
            color=end_color,
            s=end_ms**2,
            zorder=5,
            alpha=0.9,
            edgecolor="k",
        )

    # Axis formatting
    if orientation == "horizontal":
        ax.set_xlabel("Time")
        ax.set_ylabel("Lineage")
    else:
        ax.set_ylabel("Time")
        ax.set_xlabel("Lineage")
        ax.invert_yaxis()
    ax.autoscale()
    ax.set_aspect("auto")
    plt.tight_layout()

    # Optional: interactive tooltips using mplcursors
    if interactive_tooltip:
        try:
            # pylint: disable=import-outside-toplevel
            import mplcursors

            cursor = mplcursors.cursor(main_nodes, hover=True)
            cursor.connect(
                "add",
                lambda sel: sel.annotation.set_text(
                    data["hover_texts"][sel.index]
                    .replace("<br>", "\n")
                    .replace("<b>", "")
                    .replace("</b>", "")
                    .replace("<span style='font-family:monospace'>", "")
                    .replace("</span>", "")
                ),
            )
        except ImportError:
            print("mplcursors not installed; install for interactive node tooltips.")

    return fig


def plotly_cell_lineage(
    G,
    time_feature="t",
    orientation="horizontal",
    show_label=True,
    label_name=None,
    node_marker="circle",
    node_ms=10,
    line_color="blue",
    line_width=2,
    mark_births=False,
    birth_color="red",
    birth_marker=None,
    birth_ms=16,
    mark_ends=False,
    end_color="orange",
    end_marker="square",
    end_ms=14,
    figure_title="Cell Lineage",
    fig_height=500,
    fig_width=1000,
):
    """
    Plot a cell lineage tree as an interactive Plotly chart.
    Node hover shows all features as a readable (monospace) "pseudo-table".

    Parameters
    ----------
    G : nx.DiGraph
        The lineage graph.
    time_feature : str
        Node attribute for x-axis (typically time).
    orientation : str
        'horizontal' (time on x) or 'vertical' (time on y).
    show_label : bool
        Show node labels on the plot.
    label_name : str or None
        Node attribute to use for label (default: node name).
    node_marker : str
        Marker style for all nodes.
    node_ms : int
        Marker size for all nodes.
    line_color : str
        Color for edges and nodes.
    line_width : int or float
        Edge line width.
    mark_births : bool
        Mark new tracks with a special marker/color.
    birth_color : str
        Color for birth marker.
    birth_marker : str
        Marker for birth nodes.
    birth_ms : int
        Size for birth marker.
    mark_ends : bool
        Mark track ends with a special marker/color.
    end_color : str
        Color for end marker.
    end_marker : str
        Marker for end nodes.
    end_ms : int
        Size for end marker.
    figure_title : str
        Plot title.
    fig_height : str
        Height of the plotly figure.
    fig_width: int
        Width of the plotly figure.
    """
    assigned_y = compute_lineage_y(G, time_feature)
    data = extract_lineage_plotdata(
        G, assigned_y, time_feature, label_name, orientation
    )

    fig = go.Figure()

    # Draw edges as separate traces for better performance (and control)
    for x, y in zip(data["edge_xs"], data["edge_ys"]):
        fig.add_trace(
            go.Scatter(
                x=x,
                y=y,
                mode="lines",
                line=dict(color=line_color, width=line_width),
                hoverinfo="skip",
                showlegend=False,
            )
        )

    # Main node markers with pseudo-table hover (HTML with <br>, monospace)
    fig.add_trace(
        go.Scatter(
            x=data["xs"],
            y=data["ys"],
            mode="markers+text" if show_label else "markers",
            marker=dict(
                symbol=node_marker, color=line_color, size=node_ms, line=dict(width=0)
            ),
            text=data["node_labels"] if show_label else None,
            textposition="top center",
            hovertemplate="%{customdata}<extra></extra>",
            customdata=data["hover_texts"],
            name="Cells",
        )
    )

    # Markers for births and ends
    if mark_births and data["births_x"]:
        marker = (
            birth_marker
            if birth_marker
            else (">" if orientation == "horizontal" else "v")
        )
        marker_map = {
            ">": "triangle-right",
            "<": "triangle-left",
            "^": "triangle-up",
            "v": "triangle-down",
            "o": "circle",
            "s": "square",
            "d": "diamond",
            "*": "star",
            "x": "x",
            "+": "cross",
        }
        marker_symbol = marker_map.get(
            marker, "triangle-right" if orientation == "horizontal" else "triangle-down"
        )
        fig.add_trace(
            go.Scatter(
                x=data["births_x"],
                y=data["births_y"],
                mode="markers",
                marker=dict(
                    symbol=marker_symbol,
                    color=birth_color,
                    size=birth_ms,
                    line=dict(width=1, color=birth_color),
                ),
                name="Cell birth",
                hoverinfo="skip",
                showlegend=True,
            )
        )
    if mark_ends and data["ends_x"]:
        marker_map = {
            "s": "square",
            "o": "circle",
            "d": "diamond",
            "*": "star",
            "x": "x",
            "+": "cross",
            ">": "triangle-right",
            "<": "triangle-left",
            "^": "triangle-up",
            "v": "triangle-down",
        }
        marker_symbol = marker_map.get(end_marker, "square")
        fig.add_trace(
            go.Scatter(
                x=data["ends_x"],
                y=data["ends_y"],
                mode="markers",
                marker=dict(
                    symbol=marker_symbol,
                    color=end_color,
                    size=end_ms,
                    line=dict(width=1, color=end_color),
                ),
                name="Cell end",
                hoverinfo="skip",
                showlegend=True,
            )
        )

    # Axes and layout
    if orientation == "horizontal":
        fig.update_xaxes(title="Time")
        fig.update_yaxes(title="Lineage")
    else:
        fig.update_xaxes(title="Lineage")
        fig.update_yaxes(title="Time", autorange="reversed")

    fig.update_layout(
        title=figure_title, height=fig_height, width=fig_width, plot_bgcolor="white"
    )
    return fig



"""Segmenter using CellposeSAM: https://doi.org/10.1101/2025.04.28.651001"""

from cellpose import core, models

from acia.attribute import attribute_segmentation
from acia.base import ImageSequenceSource, Overlay
from acia.segm.formats import overlay_from_masks


class CellposeSAMSegmenter:
    """CellposeSAMSegmenter using Cellpose SAM: https://doi.org/10.1101/2025.04.28.651001"""

    def __init__(self, use_GPU=None):
        if use_GPU is None:
            self.use_GPU = core.use_gpu()
        print(f"Use GPU? {self.use_GPU}")

        # create CellPose model
        self.model = models.CellposeModel(gpu=self.use_GPU)

    @staticmethod
    def __predict(images, model, cellpose_params=None):
        chans = [0, 0]  # this means segment based on first channel, no second channel

        if cellpose_params is None:
            cellpose_params = {}

        # perform inference
        masks, _, _ = model.eval(images, channels=chans, **cellpose_params)

        return masks

    def __call__(self, images: ImageSequenceSource, cellpose_params=None) -> Overlay:

        # list of images
        imgs = [im.raw for im in images]

        # perform the prediction
        masks = self.__predict(imgs, self.model, cellpose_params=cellpose_params)

        # parse the overlay
        ov = overlay_from_masks(masks)

        attribute_segmentation(ov, self)

        return ov



"""Unit test cases for tracking subsampling"""

import unittest
from pathlib import Path
from shutil import rmtree

import numpy as np
import tifffile

from acia.segm.local import LocalImage
from acia.tracking.formats import read_ctc_tracking, write_ctc_tracking


class TestSubsampling(unittest.TestCase):
    """Test the linearization of z and t stacks"""

    def __test_tracking(self, ov, tracklet_graph, tracking_graph):
        self.assertEqual(len(ov), 67)

        self.assertEqual(len(tracklet_graph.nodes), 10)
        self.assertEqual(len(tracklet_graph.edges), 8)

        self.assertEqual(len(tracking_graph.nodes), 67)
        self.assertEqual(len(tracking_graph.edges), 65)

    def test_read_write(self):
        """Test subsampling of a non-dividing contour"""

        input_path = Path("test_resources") / "ctc_format"

        ov, tracklet_graph, tracking_graph = read_ctc_tracking(input_path=input_path)

        total_area = np.sum([cont.area for cont in ov])

        self.__test_tracking(ov, tracklet_graph, tracking_graph)

        output_path = Path("tracking_output")
        if output_path.exists():
            rmtree(output_path)
        output_path.mkdir()

        images = [LocalImage(tifffile.imread(input_path / "man_track0000.tif"))]

        write_ctc_tracking(output_path, images, ov, tracklet_graph)

        ov2, tracklet_graph2, tracking_graph2 = read_ctc_tracking(
            input_path=output_path
        )

        total_area2 = np.sum([cont.area for cont in ov2])

        self.assertEqual(total_area, total_area2)

        self.__test_tracking(ov2, tracklet_graph2, tracking_graph2)


if __name__ == "__main__":
    unittest.main()



""" Utils for segmentation testing """

import unittest

import numpy as np
from shapely.geometry import Polygon

import cv2

from acia.segm.processor.offline import FlexibleModel, ModelDescriptor
from acia.segm.local import LocalImageSource, LocalImage, LocalSequenceSource

import os

os.environ["CACHE_FOLDER"] = "/tmp"

class TestIndexing(unittest.TestCase):
    """Test the linearization of z and t stacks"""

    def test_both(self):

        model_desc = ModelDescriptor(
            repo="https://gitlab+deploy-token-281:TZYmjRQZzLZsBfWsd2XS@jugit.fz-juelich.de/mlflow-executors/omnipose-executor.git",
            parameters={
                # default omnipose model
                "model": "https://fz-juelich.sciebo.de/s/3J8Z7MrADMtw9fz/download"
            },
            entry_point="main",
            version="main"
        )

        fm = FlexibleModel(
            modelDesc=model_desc,
        )

        lis = LocalImageSource.from_file("lut_image.png") #LocalImageSource(LocalImage(cv2.imread("lut_image.png")))

        res = fm.predict(lis)

        print(len(res))
        


if __name__ == "__main__":
    unittest.main()



"""Unit test cases for tracking processors"""

import unittest
from pathlib import Path

import numpy as np

from acia.segm.local import InMemorySequenceSource
from acia.tracking.formats import read_ctc_tracking

# from acia.tracking.processor.laptrack import LAPTracker
from acia.tracking.processor.trackastra import TrackastraTracker


class TestTrackingProcessors(unittest.TestCase):
    """Test the linearization of z and t stacks"""

    def __init__(self, methodName="runTest"):
        super().__init__(methodName)

        # Just generate an empty image sequence here! It is not about doing a correct tracking!
        self.images = InMemorySequenceSource(np.zeros((30, 677, 426), dtype=np.uint8))

    def test_trackastra(self):
        """Test subsampling of a non-dividing contour"""

        input_path = Path("test_resources") / "ctc_format"

        ov, _, _ = read_ctc_tracking(input_path=input_path)

        tracker = TrackastraTracker()

        # Just make sure we can execute the method
        tracker(self.images, ov)


#    def test_lap(self):
#        """Test subsampling of a non-dividing contour"""
#
#        input_path = Path("test_resources") / "ctc_format"
#
#        ov, _, _ = read_ctc_tracking(input_path=input_path)
#
#        tracker = LAPTracker()
#
#        # Just make sure we can execute the method
#        tracker(self.images, ov)



"""Utilities for the visualization"""
from string import Formatter


def strfdelta(tdelta, fmt="{D:02}d {H:02}h {M:02}m {S:02}s", inputtype="timedelta"):
    """Convert a datetime.timedelta object or a regular number to a custom-
    formatted string, just like the stftime() method does for datetime.datetime
    objects.

    The fmt argument allows custom formatting to be specified.  Fields can
    include seconds, minutes, hours, days, and weeks.  Each field is optional.

    Some examples:
        '{D:02}d {H:02}h {M:02}m {S:02}s' --> '05d 08h 04m 02s' (default)
        '{W}w {D}d {H}:{M:02}:{S:02}'     --> '4w 5d 8:04:02'
        '{D:2}d {H:2}:{M:02}:{S:02}'      --> ' 5d  8:04:02'
        '{H}h {S}s'                       --> '72h 800s'

    The inputtype argument allows tdelta to be a regular number instead of the
    default, which is a datetime.timedelta object.  Valid inputtype strings:
        's', 'seconds',
        'm', 'minutes',
        'h', 'hours',
        'd', 'days',
        'w', 'weeks'
    """

    # Convert tdelta to integer seconds.
    if inputtype == "timedelta":
        remainder = int(tdelta.total_seconds())
    elif inputtype in ["s", "seconds"]:
        remainder = int(tdelta)
    elif inputtype in ["m", "minutes"]:
        remainder = int(tdelta) * 60
    elif inputtype in ["h", "hours"]:
        remainder = int(tdelta) * 3600
    elif inputtype in ["d", "days"]:
        remainder = int(tdelta) * 86400
    elif inputtype in ["w", "weeks"]:
        remainder = int(tdelta) * 604800

    f = Formatter()
    desired_fields = [field_tuple[1] for field_tuple in f.parse(fmt)]
    possible_fields = ("W", "D", "H", "M", "S")
    constants = {"W": 604800, "D": 86400, "H": 3600, "M": 60, "S": 1}
    values = {}
    for field in possible_fields:
        if field in desired_fields and field in constants:
            values[field], remainder = divmod(remainder, constants[field])
    return f.format(fmt, **values)



""" Test acia base functionality"""

import unittest

import numpy as np

import time

from acia.segm.output import fast_mask_rendering


class TestMaskRendering(unittest.TestCase):
    """Test contour functionality"""

    def test_fast_mask(self):
        mask = np.array([
            [0, 0, 0, 0, 0],
            [0, 0, 3, 3, 0],
            [0, 3, 3, 3, 3],
            [0, 2, 2, 0, 0],
        ], dtype=np.uint8)

        im = np.zeros((*mask.shape, 3), dtype=np.uint8)

        colors = [(255, 0, 0), ]

        rendered_image = fast_mask_rendering((mask==3)[None], np.copy(im), colors)
        np.testing.assert_equal(rendered_image[...,0] == 127, mask==3)

        colors = [(255, 0, 0), (0, 0, 255)]
        rendered_image = fast_mask_rendering(np.stack([(mask==3), mask==2]), np.copy(im), colors)

        np.testing.assert_equal(rendered_image[...,0] == 127, mask==3)
        np.testing.assert_equal(rendered_image[...,2] == 127, mask==2)

        #print(rendered_image)
        #print(rendered_image.shape)


if __name__ == "__main__":
    unittest.main()



"""Tracking Processor definition"""

import networkx as nx

from acia.base import ImageSequenceSource, Overlay


class TrackingProcessor:
    """Base class for tracking processors"""

    def __call__(
        self, images: ImageSequenceSource, segmentation: Overlay
    ) -> tuple[Overlay, nx.DiGraph, nx.DiGraph]:
        """Perform cell tracking

        Args:
            images (ImageSequenceSource): 2D+t image sequence
            segmentation (Overlay): Segmentation for this image sequence

        Returns:
            tuple[Overlay, nx.DiGraph, nx.DiGraph]: Returns new overlay, the tracklet graph (every cell cycle is a node) and the tracking graph (every cell detection is a node)
        """
        raise NotImplementedError()



"""Utility functions for the tracking processors"""

import numpy as np

from acia.base import Overlay


def overlay_to_masks(segmentation: Overlay, height: int, width: int):
    masks = []
    for frame_ov in segmentation.timeIterator():

        # deal with empty masks
        mask = np.zeros((height, width), dtype=np.uint8)
        if len(frame_ov) > 0:
            mask = frame_ov.toMasks(height=height, width=width, binary_mask=False)[0]

        masks.append(mask)

    # Convert to numpy inputs
    masks = np.stack(masks)

    return masks
